import csv
import editdistance
import logging
import os
import multiprocessing as mp

# suppress super-verbose nmslib logging
logging.getLogger('nmslib').setLevel(logging.WARNING)
import nmslib

import numpy as np
import itertools
import pandas as pd
import pickle
import string

import pyarrow as pa
import pyarrow.parquet as pq

from collections import defaultdict
from sklearn.feature_extraction.text import CountVectorizer

from namematch.data_structures.schema import Schema
from namematch.data_structures.parameters import Parameters
from namematch.base import NamematchBase
from namematch.utils.utils import (
    log_runtime_and_memory,
    get_nn_string_from_blockstring,
    build_blockstring,
    get_endpoints,
    get_ed_string_from_blockstring,
)
from namematch.utils.profiler import Profiler

profile = Profiler()

logger = logging.getLogger()

class Block(NamematchBase):
    '''
    Args:
        params (Parameters object): contains matching parameter values
        schema (Schema object): contains match schema info (files to match, variables to use, etc.)
        all_names_file (str): path to the all-names file
        must_links_file (str): path to the must-links file
        blocking_index_bin_file: name of blocking index file
        og_blocking_index_file (str): path to a pre-built nmslib index (optional, if doesn't exist then None)
        candidate_pairs_file (str): path to the candidate-pairs file

    '''
    def __init__(
        self,
        params,
        schema,
        all_names_file='all_names.parquet',
        must_links_file='must_links.csv',
        candidate_pairs_file='candidate_pairs.parquet',
        blocking_index_bin_file='blocking_index.bin',
        og_blocking_index_file='None',
        *args,
        **kwargs
    ):
        super(Block, self).__init__(params, schema, *args, **kwargs)

        self.all_names_file = all_names_file
        self.must_links_file = must_links_file
        self.og_blocking_index_file = og_blocking_index_file
        self.main_index_file = blocking_index_bin_file
        self.candidate_pairs_file = candidate_pairs_file

    @property
    def output_files(self):
        output_files = [
            self.candidate_pairs_file,
            self.main_index_file,
            self.main_index_file + '.pkl',
            self.main_index_file + '.dat'
        ]
        if not self.params.incremental:
            temp_dir = os.path.dirname(self.candidate_pairs_file)
            output_files.append(os.path.join(temp_dir, 'uncovered_pairs.csv'))
        return output_files

    # @log_runtime_and_memory
    @profile
    def main(self, **kw):
        '''Generate the candidate-pairs list using the blocking scheme outlined in the config.'''
        temp_dir = os.path.dirname(self.candidate_pairs_file)

        nn_cols, ed_col, absval_col = get_blocking_columns(self.params.blocking_scheme)

        # check if the absval_col is needed -- if not, set to None
        temp = pd.read_parquet(self.all_names_file, columns=['drop_from_nm', ed_col])
        if temp[temp.drop_from_nm == 0][[ed_col]].query(f"{ed_col} == '' or {ed_col}.isnull()").shape[0] == 0:
            absval_col = None

        an = read_an(self.all_names_file, nn_cols, ed_col, absval_col)

        nn_string_info, nn_string_expanded_df = \
                self.convert_all_names_to_blockstring_info(an, absval_col, self.params)

        main_index, \
        main_index_nn_strings, \
        second_index, \
        second_index_nn_strings = self.get_indices(
                self.params,
                nn_string_info.nn_string.tolist(),
                self.og_blocking_index_file)

        # only query names that appear in the new data
        nn_string_info_to_query, \
        nn_strings_to_query, \
        shingles_to_query = self.get_query_strings(nn_string_info, self.params.blocking_scheme)

        if os.path.exists(self.candidate_pairs_file):
            os.remove(self.candidate_pairs_file)

        exact_match_cp_df = self.get_exact_match_candidate_pairs(
                nn_string_info_to_query[nn_string_info_to_query.n_total > 1],
                nn_string_expanded_df,
                self.params.blocking_thresholds)
        write_some_cps(exact_match_cp_df, self.candidate_pairs_file, header=True)

        candidate_pairs_df = self.generate_candidate_pairs(
            nn_strings_to_query,
            shingles_to_query,
            nn_string_info,
            nn_string_expanded_df,
            main_index,
            main_index_nn_strings,
            second_index,
            second_index_nn_strings,
            batch_size=self.params.block_batch_size,
        )

        # evaluate blocking
        if not self.params.incremental:
            must_links_df = pd.read_csv(self.must_links_file)

            true_pairs_df = generate_true_pairs(must_links_df)
            uncovered_pairs_df = self.evaluate_blocking(
                    candidate_pairs_df, true_pairs_df, self.params.blocking_scheme)

            candidate_pairs_df = self.add_uncovered_pairs(candidate_pairs_df, uncovered_pairs_df)

            uncovered_pairs_df.to_csv( # only needed for each sanity check glance
                os.path.join(temp_dir, 'uncovered_pairs.csv'), index=False,
                quoting=csv.QUOTE_NONNUMERIC, sep=',')

        # output
        table = pa.Table.from_pandas(candidate_pairs_df)
        pq.write_table(table, self.candidate_pairs_file)

        save_main_index(main_index, main_index_nn_strings, self.main_index_file)

        if self.enable_lprof:
            self.write_line_profile_stats(profile.line_profiler)

    # @log_runtime_and_memory
    @profile
    def split_last_names(self, df, last_name_column, blocking_scheme, **kw):
        '''Expand the processed all-names file to handle double last names (e.g. SAM SMITH-BROWN
        becomes SAM SMITH and SAM BROWN).

        Args:
            df: all-names table, relevant columns only (where drop_from_nm == 0)
            last_name_column (str): clean last name column
            blocking_scheme (dict): dictionary with info on how to do blocking

        Returns:
            pd.DataFrame: more rows than input all names, plus orig_last_name and orig_record columns
        '''

        from collections import Counter

        df = df.copy()

        # preliminary cleaning

        df["orig_last_name"] = df[last_name_column]

        names_to_expand = df[df[last_name_column].str.contains(' ')].copy()

        if len(names_to_expand) > 0:

            # if there are multiple spaces, split on the last one
            names_to_expand["split_names"] = (
                    names_to_expand[last_name_column].str
                    .rsplit(" ", 1))

            # unpack the list of splits, so each is in its own row
            id_vars = names_to_expand.columns.tolist()
            id_vars.remove("split_names")
            split_up = (names_to_expand["split_names"]
                    .apply(pd.Series)
                    .merge(names_to_expand, left_index=True, right_index=True)
                    .drop(columns=["split_names"], axis=1)
                    .melt(id_vars=id_vars, value_name="split_name")
                    .drop(columns=["variable"], axis=1)
                    .dropna()
                    .drop_duplicates())

            split_up[last_name_column] = split_up["split_name"]
            split_up = split_up.drop(columns=["split_name"])

            # make sure the original name is there as well
            df = pd.concat([df, split_up])

        df = df.drop_duplicates()
        df["orig_record"] = (df[last_name_column] == df["orig_last_name"]).astype(int)
        df = df.reset_index(drop=True)

        df['nn_string_full'] = df.nn_string
        df['nn_string'] = build_blockstring(df, blocking_scheme, incl_ed_string=False)

        return df

    # @log_runtime_and_memory
    @profile
    def convert_all_names_to_blockstring_info(self, an, absval_col, params, **kw):
        '''Create a table with information about blockstrings. If the split_names parameter is True,
        then this function expands double last names to create two new "records" (e.g. SAM SMITH-BROWN
        becomes SAM SMITH and SAM BROWN).

        Args:
            an (pd.DataFrame): all-names table, relevant columns only (where drop_from_nm == 0)

                =======================  =======================================================
                record_id                unique record identifier
                blockstring              concatenation of all the blocking variables (sep by ::)
                file_type                either "new" or "existing"
                drop_from_nm             flag, 1 if met any "to drop" criteria 0 otherwise
                <nn-blocking column(s)>  variables for near-neighbor blocking
                <ed-blocking column>     variable for edit-distance blocking
                <av-blocking column>     (optional) variable for abs-value blocking
                nn_string                concatenated version of nn-blocking columns (sep by ::)
                ed_string                copy of ed-blocking column
                absval_string            copy of abs-value-blocking column
                =======================  =======================================================

            absval_col (str): column for absolute-value blocking
            params (Parameter object): contains matching parameters

        Returns:
                tuple: tuple containing:

                - **nn_string_info** (*pd.DataFrame*): table with one row per nn_string (or expanded nn_string)

                    ======================   ==============================================================
                    nn_string                concatenated version of nn-blocking columns (sep by ::)
                    commonness_penalty       float indicating how common the last name is
                    n_new                    number of times this nn_string appears in a "new" record
                    n_existing               number of times this nn_string appears in an "existing" record
                    n_total                  number of times this nn_string appears in any record
                    ======================   ==============================================================

                - **nn_string_expanded_df** (*pd.DataFrame*): table with one row per blockstring (or expanded blockstring)

                    ======================   ========================================================================
                    nn_string                concatenated version of nn-blocking columns (sep by ::)
                    nn_string_full           (optional) if split_names is True, this is the full (un-split) nn_string
                    ed_string                copy of ed-blocking column
                    absval_string            copy of abs-value-blocking column
                    ======================   ========================================================================

        '''


        an = an.copy()

        logger.info("Preparing blockstring data.")

        expanded_cols = ['nn_string']
        an_use = an.copy()

        if params.split_names:
            logger.trace('Splitting last names.')
            an_split_ln = self.split_last_names(an, params.last_name_column, params.blocking_scheme)
            an_use = an_split_ln.copy()
            expanded_cols.append('nn_string_full')

        expanded_cols.append('ed_string')
        if absval_col is not None:
            expanded_cols.append('absval_string')

        nn_string_expanded_df = an_use[expanded_cols].drop_duplicates().set_index('nn_string')

        # create dictionary mapping nn_strings to their record counts
        # NOTE: dict has two keys, "new" and "old"
        nn_strings_to_num_recs = get_nn_string_counts(an_use)

        # count last name frequency to establish a measure of how common a name is
        commonness_penalty_lookup = get_common_name_penalties(
                an[params.last_name_column],
                params.blocking_thresholds['common_name_max_penalty'])
        an_use['commonness_penalty'] = an_use[params.last_name_column].map(commonness_penalty_lookup)
        # NOTE: should the params.last_name_column in the above line be original last name always instead?

        nn_string_info = an_use[['nn_string', 'commonness_penalty']].drop_duplicates()

        if len(nn_string_info) != nn_string_info.nn_string.nunique():
            logger.error(f'More than one last name tied to nn_string, meaning there are '
                         f'multiple commonness_penalties per nn_string. Will cause a weird merge.')
            raise

        nn_string_info['n_new'] = nn_string_info.nn_string.map(nn_strings_to_num_recs['new'])
        nn_string_info['n_existing'] = nn_string_info.nn_string.map(nn_strings_to_num_recs['existing'])
        nn_string_info['n_total'] = nn_string_info.n_new + nn_string_info.n_existing

        return nn_string_info, nn_string_expanded_df


    def get_query_strings(self, nn_string_info, blocking_scheme):
        '''Filter to nn_strings that appear in the new data -- these are the only strings
        for which we need near-neighbors. If incremental is False, this filtering step does nothing.

        Args:
            nn_string_info (pd.DataFrame): table with one row per nn_string (or expanded nn_string)

                ======================   ==============================================================
                nn_string                concatenated version of nn-blocking columns (sep by ::)
                commonness_penalty       float indicating how common the last name is
                n_new                    number of times this nn_string appears in a "new" record
                n_existing               number of times this nn_string appears in an "existing" record
                n_total                  number of times this nn_string appears in any record
                ======================   ==============================================================

            blocking_scheme (dict): dictionary with info on how to do blocking

        Returns:
                tuple: tuple containing:

                - **nn_string_info_to_query** (*pd.DataFrame*): nn_string_info, subset to nn_strings where n_new > 0
                - **nn_strings_to_query** (*list*): nn_strings that appear at least once in a "new" record
                - **shingles_to_query** (*scipy.sparse.csr_matrix*): sparse weighted shingles matrix for the nn_strings that appear in a new record
        '''

        nn_string_info = nn_string_info.copy()

        nn_string_info_to_query = nn_string_info[nn_string_info.n_new > 0].copy()
        nn_strings_to_query = nn_string_info_to_query.nn_string.tolist()

        shingles_to_query = self.generate_shingles_matrix(
            nn_strings_to_query,
            blocking_scheme['alpha'],
            blocking_scheme['power'],
            matrix_type='query')

        return nn_string_info_to_query, nn_strings_to_query, shingles_to_query

    # @log_runtime_and_memory
    @profile
    def generate_shingles_matrix(
            self,
            nn_strings,
            alpha,
            power,
            matrix_type,
            verbose=True,
            **kw):
        '''Return a weighted sparse matrix of 2-shingles

        Args:
            alpha (float): weight of LAST relative to FIRST
            power (float): parameter controlling the impact of name length on cosine distance
            matrix_type (str): description of matrix being built (for logging)
            verbose (bool): True if status messages desired

        Returns:
            scipy.sparse.csr_matrix: Weighted sparse 2-shingles matrix
        '''

        nn_strings = nn_strings[:]

        all_shingles = get_all_shingles()

        if verbose:
            logger.info(f"Creating {matrix_type} shingles matrix.")

        # shingles input strings
        vectorizer = CountVectorizer(
            analyzer='char',
            ngram_range=(2,2),
            dtype=np.double,
            decode_error='replace',
            vocabulary=all_shingles,
            lowercase=False)

        def make_matrix(i, num):
            formatted = ['*' + nns.split('::')[i] + '*' for nns in nn_strings]
            shingles_matrix = vectorizer.fit_transform(formatted)
            weights = num / np.power(shingles_matrix.sum(axis=1), power)
            shingles_matrix = shingles_matrix.multiply(weights)
            return shingles_matrix

        # sum of shingles for the first name should equal 1
        var1_shingles_matrix = make_matrix(0, 1)
        # sum of shingles for the last name should equal alpha
        var2_shingles_matrix = make_matrix(1, alpha)
        # combine weighted matrices
        shingles_matrix = var1_shingles_matrix + var2_shingles_matrix

        return shingles_matrix

    # @log_runtime_and_memory
    @profile
    def load_main_index(self, index_file, **kw):
        '''Load the main index, which is reusable over time as data is added incrementally.

        Args:
            index_file (str): path to stored index

        Returns:
            nmslib.FloatIndex: nmslib index object
        '''

        ix = prep_index()
        ix.loadIndex(index_file, load_data=True)

        return ix

    # @log_runtime_and_memory
    def generate_index(
            self,
            nn_strings,
            num_workers,
            M,
            efC,
            post,
            alpha,
            power,
            print_progress=True,
            **kw):
        '''Build an nmslib index based on a list of nn_strings and a set of parameters.

        Args:
            nn_strings (list): strings of the form 'FIRST::LAST' to shingle and put in matrix (rows)
            num_workers (int): number of threads nmslib should use when parallelizing
            M, efc, post: nmslib parameters
            alpha (float): weight of last-name relative to first-name
            power (float): parameter controlling the impact of name length on cosine distance
            print_progress (bool): controls verbosity of index creation

        Returns:
            nmslib.FloatIndex: nmslib index object
        '''

        index_params = {
            'M' : M,
            'indexThreadQty' : num_workers,
            'efConstruction' : efC,
            'post' : post
        }

        shingles_matrix = self.generate_shingles_matrix(nn_strings, alpha, power, 'index')
        ix = prep_index()
        # batch load data points
        temp = ix.addDataPointBatch(
            data=shingles_matrix,
            ids=np.arange(shingles_matrix.shape[0], dtype=np.int32))
        ix.createIndex(index_params, print_progress=print_progress)

        return ix

    # @log_runtime_and_memory
    @profile
    def get_indices(self, params, all_nn_strings, og_blocking_index_file, **kw):
        '''Wrapper function coordinating the creation and/or loading of the nmslib indices.

        Args:
            params (Parameters object): contains matching parameter values
            all_nn_strings: list of all unique nn_strings in the data (expanded if split_names is True)
            og_blocking_index_file: path to a pre-build nmslib index (optional, if doesn't exist then None)

        Returns:
               tuple: tuple containing:

                - **main_index** (*nmslib.FloatIndex*): the main nmslib index
                - **main_index_nn_strings** (*list*): nn_strings that are in the main nmslib index
                - **second_index** (*nmslib.FloatIndex*): the secondary nmslib index for querying new nn_strings during incremental runs (often None)
                - **second_index_nn_strings** (*list*): nn_strings that are in the secondary nmslib index (often None)
        '''


        main_index = None
        second_index = None
        main_index_nn_strings = None
        second_index_nn_strings = None

        build_index_from_scratch = ((not params.incremental) or \
                                    og_blocking_index_file == 'None' or \
                                    params.index['rebuild_main_index'] == 'always')

        # if not build_index_from_scratch so far, check one more time consuming trigger
        if (not build_index_from_scratch) and params.index['rebuild_main_index'] == 'if_secondary_index_exceeds_limit':
            main_index_nn_strings = load_main_index_nn_strings(og_blocking_index_file)
            second_index_nn_strings = get_second_index_nn_strings(all_nn_strings, main_index_nn_strings)
            if len(second_index_nn_strings) >= params.index['secondary_index_limit']:
                build_index_from_scratch = True

        if build_index_from_scratch:

            logger.trace('Building index from scratch.')
            main_index = self.generate_index(
                    all_nn_strings,
                    params.num_workers, params.nmslib['M'], params.nmslib['efC'],
                    params.nmslib['post'], params.blocking_scheme['alpha'],
                    params.blocking_scheme['power'])
            main_index_nn_strings = all_nn_strings[:]

        else:

            if second_index_nn_strings is None:

                main_index_nn_strings = load_main_index_nn_strings(og_blocking_index_file)
                second_index_nn_strings = get_second_index_nn_strings(all_nn_strings, main_index_nn_strings)

            # load main index
            logger.trace('Loading main index.')
            main_index = self.load_main_index(og_blocking_index_file)

            # build second index
            if len(second_index_nn_strings) > 0:

                logger.trace('Building second index.')
                second_index = self.generate_index(
                        second_index_nn_strings,
                        params.num_workers, params.nmslib['M'], params.nmslib['efC'],
                        params.nmslib['post'], params.blocking_scheme['alpha'],
                        params.blocking_scheme['power'])

        return main_index, main_index_nn_strings, second_index, second_index_nn_strings

    # @log_runtime_and_memory
    @profile
    def generate_candidate_pairs(
        self,
        nn_strings_to_query,
        shingles_to_query,
        nn_string_info,
        nn_string_expanded_df,
        main_index,
        main_index_nn_strings,
        second_index,
        second_index_nn_strings,
        batch_size,
        **kw
    ):
        '''Wrapper function for querying the nmslib index (or indices) and getting
        non-matching candidate pairs.

        Args:
            nn_strings_to_query (list): nn_strings in new data -- those that need near neighbors
            shingles_to_query (csr_matrix): shingles matrix for nn_strings_to_query
            nn_string_info (pd.DataFrame):  table with one row per nn_string (or expanded nn_string)
            nn_string_expanded_df (pd.DataFrame): maps a nn_string to a ed_string and absval_string
            main_index (nmslib index): the main nmslib index for querying
            main_index_nn_strings (list): nn_strings in main_index
            second_index (nmslib index): the secondary nmslib index, for some incremental runs
            second_index_nn_strings (list): nn_strings in second_index
            batch_size (int): batch size. Default is 10000 and can be modify in config.yaml file.
        Returns:
            pd.DataFrame: candidate-pairs list, before adding in uncovered pairs

            ======================   =======================================================
            blockstring_1            concatenated version of blocking columns for first element in pair (sep by ::)
            blockstring_2            concatenated version of blocking columns for second element in pair (sep by ::)
            cos_dist                 approximate cosine distance between two nn_strings (nmslib)
            edit_dist                number of character edits between ed-strings
            covered_pair             flag; 1 for pairs that made it through blocking, 0 otherwise; all 1s here
            ======================   =======================================================
        '''
        logger.info('Querying to get candidate pairs.')

        nn_string_info = nn_string_info.copy()

        # determine which indices to query
        indices_to_query = ['main']
        if second_index is not None:
            indices_to_query.append('secondary')

        logger.info(f'Indices to query: {len(indices_to_query)}')

        # how many names do we want to query at a time?
        # runtime/ram tradeoff; can lower if having RAM issues
        # batch_size = 2000 # NOTE can be parameterized/reduced if RAM is issue
        logger.debug(f"batch size: {batch_size}")

        start_ix_batch = 0
        while start_ix_batch < len(nn_strings_to_query):
            end_ix_batch = min(start_ix_batch + batch_size, len(nn_strings_to_query))

            cand_pair_df_list = []

            if self.params.verbose is not None and (start_ix_batch % self.params.verbose == 0):
                logger.info("  %s out of %s nn_strings queried." % (
                        min(start_ix_batch, len(nn_strings_to_query)), len(nn_strings_to_query)))

            for index_type in indices_to_query:

                if index_type == 'main':
                    index_to_query = main_index
                    nn_strings_this_index = main_index_nn_strings[:]
                elif index_type == 'secondary':
                    index_to_query = second_index
                    nn_strings_this_index = second_index_nn_strings[:]
                else:
                    logger.error('Variable "index_type" must be "main" or "secondary."')
                    raise ValueError

                # query the nmslib index:
                # pass in a group of names, and get back the k most similar names)
                near_neighbors_list = index_to_query.knnQueryBatch(
                    queries=shingles_to_query[start_ix_batch:end_ix_batch],
                    k=self.params.nmslib['k'],
                    num_threads=self.params.num_workers)

                nn_strings_queried_this_batch = nn_strings_to_query[start_ix_batch:end_ix_batch]

                near_neighbors_df = self.get_near_neighbors_df(
                        near_neighbors_list,
                        nn_string_info,
                        nn_strings_this_index,
                        nn_strings_queried_this_batch)

                # parallelize the paring down for possible candidate pairs to actual candidate pairs
                end_points = get_endpoints(len(near_neighbors_df), self.params.num_workers)
                if self.params.parallelize:

                    output = mp.Queue()
                    jobs = [
                        mp.Process(
                            target = self.get_actual_candidates,
                            args = (
                                near_neighbors_df.iloc[start_ix_worker : end_ix_worker],
                                nn_string_expanded_df,
                                nn_strings_to_query,
                                self.params.blocking_thresholds,
                                self.params.incremental,
                                output)) for start_ix_worker, end_ix_worker in end_points]
                    for job in jobs:
                        job.start()

                    cand_pairs_to_add_df = [output.get() for job in jobs]

                    [job.join() for job in jobs]
                    failure_occurred = sum([job.exitcode != 0 for job in jobs])
                    if failure_occurred:
                        logger.error("Error occurred in %s worker(s)." % failure_occurred)
                        raise Exception("Error occurred in %s worker(s)." % failure_occurred)

                else:

                    cand_pairs_to_add_df = []
                    for start_ix_worker, end_ix_worker in end_points:
                        cand_pairs_to_add_df_this_iter = self.get_actual_candidates(
                                near_neighbors_df.iloc[start_ix_worker : end_ix_worker],
                                nn_string_expanded_df,
                                nn_strings_to_query,
                                self.params.blocking_thresholds,
                                self.params.incremental)
                        cand_pairs_to_add_df.append(cand_pairs_to_add_df_this_iter)

                cand_pair_df_list.extend(cand_pairs_to_add_df)

            start_ix_batch += batch_size

            cand_pair_df = pd.concat(cand_pair_df_list)
            cand_pair_df = cand_pair_df.drop_duplicates(subset=['blockstring_1', 'blockstring_2'])
            # because same name might be in both indices

            # update total number candidate pairs generated so far; write to file
            write_some_cps(cand_pair_df, self.candidate_pairs_file)
            del cand_pair_df

        # have to load so we can drop duplicates at global level...
        # because of split names
        logger.trace('Loading candidate pairs to drop duplicates.')
        cand_pair_df = pd.read_csv(self.candidate_pairs_file)
        cand_pair_df = cand_pair_df.drop_duplicates(subset=['blockstring_1', 'blockstring_2'])
        cand_pair_df.to_csv(
                self.candidate_pairs_file,
                index=False,
                quoting=csv.QUOTE_NONNUMERIC)

        return cand_pair_df

    # @log_runtime_and_memory
    def compute_cosine_sim(self, blockstrings_in_pairs, pairs_df, shingles_matrix, **kw):
        '''Fast cosine similarity computation using the shingles matrix.

        Args:
            blockstrings_in_pairs (list): used to get index of different strings in shingles_matrix
            pairs_df (pd.DataFrame): blockstrings you want cosine distance between

               ===================   =======================================================
               blockstring_1         blockstring for the first record in the pair
               blockstring_2         blockstring for the second record in the pair
               covered_pair          flag, 1 if covered 0 otherwise
               nn_strings_1          nn_string for the first record in the pair
               nn_strings_2          nn_string for the second record in the pair
               both_nn_strings       nn_string_1 + ' ' + nn_string_2
               ===================   =======================================================

        Returns:
            shingles_matrix (csr_matrix): weighted shingles matrix

        '''

        name_index_map = {}
        for i, the_name in enumerate(blockstrings_in_pairs):
            name_index_map[the_name] = i

        idx1 = []
        idx2 = []
        for i, row in pairs_df.iterrows():
            n1 = row['blockstring_1']
            n2 = row['blockstring_2']
            if n1 in name_index_map and n2 in name_index_map:
                idx1.append(name_index_map[n1])
                idx2.append(name_index_map[n2])

        mat1 = shingles_matrix[np.array(idx1)]
        mat2 = shingles_matrix[np.array(idx2)]

        numerator = mat1.multiply(mat2).sum(1)

        mat1_squared = mat1.multiply(mat1).sum(1)
        mat2_squared = mat2.multiply(mat2).sum(1)
        denom = np.multiply(np.sqrt(mat1_squared), np.sqrt(mat2_squared))

        pair_cos = 1.0 - np.divide(numerator, denom)
        logger.info('Computing cosine similarities.')

        pair_cos = pair_cos.flatten().tolist()[0]

        return pair_cos

    # @log_runtime_and_memory
    @profile
    def evaluate_blocking(self, cp_df, tp_df, blocking_scheme, **kw):
        '''The evaluate_blocking function computes the pair completeness metrics to
        determine how successful blocking was at minimizing comparisons and
        maximizing true positives (i.e. generating a candidate pair between
        records that are actually matches).

        Args:
            cp_df (pd.DataFrame): candidate pairs df
            tp_df (pd.DataFrame): true pairs df (blockstring_1, blockstring_2)
            blocking_scheme (dict): blocking_scheme (dict): dictionary with info on how to do blocking

        Returns:
            pd.DataFrame: portion of candidate-pairs dataframe where covered == 0
        '''

        cp_df = cp_df.copy()

        cp_df['nn_string_1'] = np.vectorize(
                get_nn_string_from_blockstring, otypes=['object'])(cp_df.blockstring_1)
        cp_df['nn_string_2'] = np.vectorize(
                get_nn_string_from_blockstring, otypes=['object'])(cp_df.blockstring_2)
        cp_df['both_blockstrings'] = \
                cp_df.blockstring_1 + ' ' + cp_df.blockstring_2
        cp_df['both_nn_strings'] = \
                cp_df.nn_string_1 + ' ' + cp_df.nn_string_2

        # covered/uncovered pairs (get cosine distribution)
        # -----------------------------------------------

        pasted = tp_df["blockstring_1"] + " " + tp_df["blockstring_2"]
        tp_df["covered_pair"] = pasted.isin(cp_df["both_blockstrings"]).astype(int)

        tp_df['nn_string_1'] = np.vectorize(
                get_nn_string_from_blockstring, otypes=['object'])(tp_df.blockstring_1)
        tp_df['nn_string_2'] = np.vectorize(
                get_nn_string_from_blockstring, otypes=['object'])(tp_df.blockstring_2)
        tp_df['both_nn_strings'] = \
                tp_df.nn_string_1 + ' ' + tp_df.nn_string_2

        tp_df_nonmatching = \
                tp_df[tp_df.blockstring_1 != tp_df.blockstring_2].copy()

        # calculate cosine distance between true pairs (non-matching)
        blockstrings_in_true_pairs = \
                tp_df.blockstring_1.append(
                tp_df.blockstring_2, ignore_index=True).drop_duplicates().tolist()
        tp_shingles_matrix = self.generate_shingles_matrix(
                blockstrings_in_true_pairs,
                blocking_scheme['alpha'], blocking_scheme['power'],
                matrix_type='true pairs', verbose=False)
        tp_df_nonmatching['cos_dist'] = self.compute_cosine_sim(blockstrings_in_true_pairs,
                tp_df_nonmatching, tp_shingles_matrix)

        # get the distribution of cosine distances in non-matching true pairs
        tp_nonmatching_cos_distr = pd.value_counts(
                pd.cut(tp_df_nonmatching.cos_dist, np.arange(0, 1.1, .1)),
                normalize=True, sort=False)
        logger.trace(f'Cosine distribution of non-matching true pairs: \n{tp_nonmatching_cos_distr.to_string()}')
        self.stats_dict['tp_cosine_distribution'] = tp_nonmatching_cos_distr.tolist()

        # output uncovered pairs so we can see where blocking fails
        up_df = tp_df_nonmatching[tp_df_nonmatching.covered_pair == 0].copy()
        up_df['ed_string_1'] = np.vectorize(
                get_ed_string_from_blockstring, otypes=['object'])(up_df.blockstring_1)
        up_df['ed_string_2'] = np.vectorize(
                get_ed_string_from_blockstring, otypes=['object'])(up_df.blockstring_2)
        up_df['edit_dist'] = np.vectorize(editdistance.eval, otypes=['float'])(
                up_df.ed_string_1.values, up_df.ed_string_2.values)
        up_df = up_df[['blockstring_1', 'blockstring_2', 'cos_dist',
                                                 'edit_dist', 'covered_pair']].copy()

        logger.info(f"Number of uncovered pairs: {len(up_df)}")
        self.stats_dict['n_uncovered_pairs'] = len(up_df)
        logger.info(f"Number of true pairs: {len(tp_df)}")
        self.stats_dict['n_true_blockstring_pairs'] = len(tp_df)

        # get the distribution of cosine distances in uncovered pairs
        # NOTE: fine that coming from non-matching df because no uncovered pairs will ever match
        uncovered_pair_cos_distr = pd.value_counts(
                pd.cut(up_df.cos_dist, np.arange(0, 1.1, .1)),
                normalize=True, sort=False)
        logger.trace(f'Cosine distribution of uncovered pairs: \n{uncovered_pair_cos_distr.to_string()}')
        self.stats_dict['up_cosine_distribution'] = uncovered_pair_cos_distr.tolist()

        # pair completeness, aka pc (high is good, max 1)
        # -----------------------------------------------
        # this is the proportion of actual matches that make it through blocking

        logger.info("Calculating pair completeness.")

        # calculate pair completeness for all candidate pairs
        pc_numerator_nn_string_level = tp_df[tp_df.both_nn_strings.isin(
                cp_df.both_nn_strings) == False].both_nn_strings.nunique()
        pc_numerator_blockstring_level = len(up_df)
        pc_denominator_nn_string_level = tp_df.both_nn_strings.nunique()
        pc_denominator_blockstring_level = len(tp_df)
        try:
            pc_nn_string_level = \
                    1 - (pc_numerator_nn_string_level / pc_denominator_nn_string_level)
            pc_blockstring_level = \
                    1 - (pc_numerator_blockstring_level / pc_denominator_blockstring_level)
            logger.info(f"Pair completeness, including equal blockstrings (cosine level): {round(pc_nn_string_level, 3)}")
            self.stats_dict['pc_eq_cos'] = round(pc_nn_string_level, 3)
            logger.info(f"Pair completeness, including equal blockstrings (cosine + editdistance level): {round(pc_blockstring_level, 3)}")
            self.stats_dict['pc_eq_cosed'] = round(pc_blockstring_level, 3)
        except:
            logger.warning(f"Pair completeness cannot be calculated. Make sure necessary "
                           f"files are populated.")

        # calculate pair completeness for candidate pairs that are not the
        # same (e.g. NameX, NameX) since those should be perfect
        pc_denominator_nonequal_nn_string_level = \
                tp_df[tp_df.nn_string_1 != tp_df.nn_string_2].both_nn_strings.nunique()
        pc_denominator_nonequal_blockstring_level = len(tp_df_nonmatching)
        try:
            pc_nn_string_level = \
                    1 - (pc_numerator_nn_string_level / pc_denominator_nonequal_nn_string_level)
            pc_blockstring_level = \
                    1 - (pc_numerator_blockstring_level / pc_denominator_nonequal_blockstring_level)
            logger.info(f"Pair completeness, non-equal blockstrings (cosine level): {round(pc_nn_string_level, 3)}")
            self.stats_dict['pc_neq_cos'] = round(pc_nn_string_level, 3)
            logger.info(f"Pair completeness, non-equal blockstrings (cosine + editdistance level): {round(pc_blockstring_level, 3)}")
            self.stats_dict['pc_neq_cosed'] = round(pc_blockstring_level, 3)
        except:
            logger.warning(f"Pair completeness (non-equal blockstrings) cannot be calculated. "
                           f"Make sure necessary files are populated.")

        logger.info("Finished evaluating blocking.")

        return up_df

    def add_uncovered_pairs(self, candidate_pairs_df, uncovered_pairs_df):
        '''Add the uncovered pairs to the candidate pairs dataframe so that all of the
        known pairs are in the candidate pairs list.

        Args:
            candidate_pairs_df (pd.DataFrame): candidate pairs file produced by blocking
            uncovered_pairs_df (pd.DataFrame): uncovered pairs produced by evaluating blocking

        Return:
            pd.DataFrame: candidate-pairs file

            ======================   =======================================================
            blockstring_1            concatenated version of blocking columns for first element in pair (sep by ::)
            blockstring_2            concatenated version of blocking columns for second element in pair (sep by ::)
            cos_dist                 approximate cosine distance between two nn_strings (nmslib)
            edit_dist                number of character edits between ed-strings
            covered_pair             flag; 1 for pairs that made it through blocking, 0 otherwise
            ======================   =======================================================
        '''

        logger.info(f'Blocked covered pairs: {len(candidate_pairs_df)}')
        self.stats_dict['n_bcp'] = len(candidate_pairs_df)
        logger.info(f'Uncovered pairs: {len(uncovered_pairs_df)}')
        self.stats_dict['n_uncovered_pairs'] = len(uncovered_pairs_df)

        up_cols = [col for col in uncovered_pairs_df.columns.tolist() if col in candidate_pairs_df]
        candidate_pairs_df = pd.concat([
                candidate_pairs_df,
                uncovered_pairs_df[up_cols]], ignore_index=True, sort=False)
        candidate_pairs_df = candidate_pairs_df.sample(frac=1)

        self.stats_dict['n_cand_pairs'] = len(candidate_pairs_df)
        return candidate_pairs_df

    @profile
    def apply_blocking_filter(self, df, thresholds, nn_string_expanded_df, nns_match=False):
        '''Compare similarity of names and DOBs to see if a pair of records are likely to be a match.

        Args:
            df (pd.DataFrame): holds similarity and commonness info about pairs of names

                ======================   ==================================================================================
                nn_string_1              concatenated version of nn-blocking columns for first element in pair (sep by ::)
                nn_string_2              concatenated version of nn-blocking columns for second element in pair (sep by ::)
                cos_dist                 approximate cosine distance between two nn_strings (nmslib)
                commonness_penalty_1     penalty for last-name commonness for first element in pair
                commonness_penalty_2     penalty for last-name commonness for second element in pair
                ======================   ==================================================================================

            thresholds (dict): information about what blocking distances are allowed
            nn_string_expanded_df (pd.DataFrame): maps a nn_string to a ed_string and absval_string
            nns_match (bool): True if this function is called by get_exact_match_candidate_pairs

        Returns:
            pd.DataFrame: chunk of the candidate-pairs list

            ==============   ===============================================================================
            blockstring_1    concatenated version of blocking columns for first element in pair (sep by ::)
            blockstring_2    concatenated version of blocking columns for second element in pair (sep by ::)
            cos_dist         approximate cosine distance between two nn_strings (nmslib)
            edit_dist        number of character edits between ed-strings
            covered_pair     flag; 1 for pairs that made it through blocking, 0 otherwise; all 1s here
            ==============   ===============================================================================
        '''
        # add commonness penalty
        df['commonness_penalty'] = \
                df[['commonness_penalty_1', 'commonness_penalty_2']].mean(axis=1)
        df = df.drop(columns=['commonness_penalty_1', 'commonness_penalty_2'])

        # expand to full-name/dob/age(?) level
        df = pd.merge(df, nn_string_expanded_df, left_on='nn_string_1', right_index=True)
        df = pd.merge(df, nn_string_expanded_df, left_on='nn_string_2', right_index=True, suffixes=['_1', '_2'])
        df = df.reset_index(drop=True)
        if 'nn_string_full' in nn_string_expanded_df:
            df['nn_string_1'] = df.nn_string_full_1 # swap nn_string for nn_string_full
            df['nn_string_2'] = df.nn_string_full_2
            df = df.drop(columns=['nn_string_full_1', 'nn_string_full_2'])

        if nns_match:
            df = df[df.ed_string_1 <= df.ed_string_2]

        # use dob columns to calculate edit distance
        # initialize as either 0 (if equal and not null) or -1
        df['edit_dist'] = ((df.ed_string_1 != '') & (df.ed_string_1 == df.ed_string_2)) - 1
        where_calc_ed = ((df.ed_string_1 != '') & (df.ed_string_2 != '') & (df.ed_string_1 != df.ed_string_2))
        df.loc[where_calc_ed, 'edit_dist'] = \
                np.vectorize(editdistance.eval, otypes=['float'])(
                    df[where_calc_ed].ed_string_1.values, df[where_calc_ed].ed_string_2.values)

        df['absval_diff'] = 0
        if 'absval_string_1' in df.columns.tolist():
            # need to collapse to name/age level
            gb_cols = [col for col in df.columns.tolist() if 'absval' not in col]
            df['absval_diff'] = np.abs(df.absval_string_1 - df.absval_string_2)
            df = df.groupby(gb_cols).absval_diff.min().reset_index()

        # limit to the pairs that are most likely to be the same person
        pass_high_bar = (df.cos_dist <= (thresholds['high_cosine_bar'] - df.commonness_penalty)) & \
                        (df.edit_dist >= 0) & \
                        (df.edit_dist <= thresholds['low_editdist_bar'])

        pass_low_bar =  (df.cos_dist <= (thresholds['low_cosine_bar'] - df.commonness_penalty)) & \
                        (df.edit_dist >= 0) & \
                        (df.edit_dist <= thresholds['high_editdist_bar'])

        pass_nodob_bar = (df.cos_dist <= (thresholds['nodob_cosine_bar'] - df.commonness_penalty)) & \
                         (df.edit_dist == -1) & \
                         ((df.absval_diff <= thresholds['absvalue_bar']) | (df.absval_diff.isnull()))

        df = df[(pass_high_bar | pass_low_bar | pass_nodob_bar)].copy()

        # flip any strings that are not in alphabetical order
        # NOTE: necessary for incremental runs
        df['nns1_is_min'] = (df.nn_string_1.astype(str) + '::') <= (df.nn_string_2.astype(str) + '::')
        df['blockstring_1'] = df.nn_string_1.astype(str) + '::' + df.ed_string_1.astype(str)
        df['blockstring_2'] = df.nn_string_2.astype(str) + '::' + df.ed_string_2.astype(str)
        df.loc[df.nns1_is_min == False, 'blockstring_1'] = df.nn_string_2.astype(str) + '::' + df.ed_string_2.astype(str)
        df.loc[df.nns1_is_min == False, 'blockstring_2'] = df.nn_string_1.astype(str) + '::' + df.ed_string_1.astype(str)

        cand_pairs = df[['blockstring_1', 'blockstring_2', 'cos_dist', 'edit_dist']].copy()

        # once we've applied all the filters, the rows that are left will make it through blocking
        cand_pairs['covered_pair'] = 1

        return cand_pairs


    @profile
    def disallow_switched_pairs(self, df, incremental, nn_strings_to_query):
        '''Look through the columns nn_string_1 and nn_string_2 and keep only rows where
        nn_string1 <= nn_string2 to prevent duplicates in the end (i.e. ABBY->ZABBY &
        ZABBY->ABBY; only one is needed). Special case for incremental runs.

        Args:
            df (pd.DataFrame): holds similarity and commonness info about pairs of names
            incremental (bool) : Ture if current run incremental
            nn_strings_to_query (list): nn_strings that are in "to query" list

        Returns:
            pd.DataFrame: same as input df, but no AB/BA duplicates
        '''

        df = df.copy()

        # hack to help make sure things are ordered properly
        # e.g. JOHN SMITH JR > JOHN SMITH, but # e.g. JOHN SMITH JR:: < JOHN SMITH::
        # e.g. JOHN SMITH-JONES > JOHN SMITH, but # e.g. JOHN SMITH-JONES:: < JOHN SMITH::
        df['nn_string_1_'] = df.nn_string_1 + '::'
        df['nn_string_2_'] = df.nn_string_2 + '::'

        df['allowed'] = 0
        df.loc[df.nn_string_1_ < df.nn_string_2_, 'allowed'] = 1
        df.loc[(incremental) &
              (df.nn_string_1_ > df.nn_string_2_) &
              (df.nn_string_2.isin(nn_strings_to_query) == False), 'allowed'] = 1
        df = df[df.allowed == 1]
        df = df.drop(columns=['nn_string_1_', 'nn_string_2_'])

        return df

    @profile
    def get_actual_candidates(
            self,
            near_neighbors_df,
            nn_string_expanded_df,
            nn_strings_to_query,
            thresholds,
            incremental,
            output=None):
        '''Actually determines whether two names become candidates; this function is launched by
        generate_candidate_pairs() and run on individual worker threads to speed up processing.

        Args:

            near_neighbors_df (pd.DataFrame): holds similarity and commonness info about pairs of names

                ======================   =======================================================
                nn_string_ix             a string with nn_string_ix = i is the string located at nn_strings_queried_this_batch[i]
                nn_string_1              concatenated version of nn-blocking columns for first element in pair (sep by ::)
                nn_string_2              concatenated version of nn-blocking columns for second element in pair (sep by ::)
                cos_dist                 approximate cosine distance between two nn_strings (nmslib)
                commonness_penalty_1     penalty for last-name commonness for first element in pair
                commonness_penalty_2     penalty for last-name commonness for second element in pair
                ======================   =======================================================

            nn_string_expanded_df (pandas dataframe): table at nn_string/ed_string/absval_string level (expanded if split_name is True)
            nn_strings_to_query (list): nn_strings in the "to query" list (needed for incremental check)
            thresholds (dict): information about what blocking distances are allowed
            incremental (bool) : True if current run incremental
            output: None if the output should be returned, rather than written

        Returns:
            pd.DataFrame: chunk of the candidate-pairs list

            ======================   =======================================================
            blockstring_1            concatenated version of blocking columns for first element in pair (sep by ::)
            blockstring_2            concatenated version of blocking columns for second element in pair (sep by ::)
            cos_dist                 approximate cosine distance between two nn_strings (nmslib)
            edit_dist                number of character edits between ed-strings
            covered_pair             flag; 1 for pairs that made it through blocking, 0 otherwise; all 1s here
            ======================   =======================================================
        '''

        near_neighbors_df = near_neighbors_df.copy()

        # narrow down to reasaonable cosine distance
        near_neighbors_df = near_neighbors_df[near_neighbors_df.cos_dist <= thresholds['low_cosine_bar']]

        # narrow down so we don't have duplicates (AB & BA)
        near_neighbors_df = self.disallow_switched_pairs(near_neighbors_df, incremental, nn_strings_to_query)

        # filter out low edit or cosine distances
        cand_pairs_df = self.apply_blocking_filter(near_neighbors_df, thresholds, nn_string_expanded_df)

        if output is None:
            return cand_pairs_df
        else:
            output.put(cand_pairs_df)


    @profile
    def get_near_neighbors_df(
            self,
            near_neighbors_list,
            nn_string_info,
            nn_strings_this_index,
            nn_strings_queried_this_batch):
        '''For a small batch of names (nn_strings_queried_this_batch), format a dataframe that
        enumerates every pair of (name in this batch, a near neighbor), along with information about
        similarity and commonness.

        Args:
            near_neighbors_list (list): list of (list of k IDs, list of k distances) tuples, of length batch_size
            nn_string_info (pd.DataFrame): table mapping nn_string to commonness_penalty
            nn_strings_this_index (list): nn_strings in the current index
            nn_strings_queried_this_batch (list): nn_strings in the current query batch (length batch_size),
                                                  whose neighbors are stored in near_neighbors_list

        Returns:
            pd.DataFrame: holds similarity and commonness info about pairs of names

            ====================   ========================================================================================
            nn_string_ix           a string with nn_string_ix = i is the string located at nn_strings_queried_this_batch[i]
            nn_string_1            concatenated version of nn-blocking columns for first element in pair (sep by ::)
            nn_string_2            concatenated version of nn-blocking columns for second element in pair (sep by ::)
            cos_dist               approximate cosine distance between two nn_strings (nmslib)
            commonness_penalty_1   penalty for last-name commonness for first element in pair
            commonness_penalty_2   penalty for last-name commonness for second element in pair
            ====================   ========================================================================================
        '''

        nn_strings_this_index = nn_strings_this_index[:]
        nn_strings_queried_this_batch = nn_strings_queried_this_batch[:]
        nn_string_info = nn_string_info.copy()

        nn_string_info.set_index('nn_string', inplace=True)

        nn_string_info_queried_this_batch = nn_string_info.loc[nn_strings_queried_this_batch].reset_index().copy()
        nn_string_info_this_index = nn_string_info.loc[nn_strings_this_index].reset_index().copy()

        # each item in these two columns is a list of length k
        df = pd.DataFrame(near_neighbors_list, columns=['near_neighbor_nn_string_ix', 'cos_dist'])
        df['nn_string_ix'] = df.index

        # extract the k neighbor IDs, and k distances from the lists, so each is in its own cell
        # a string with nn_string_ix = i is the string located at nn_strings_queried_this_batch[i]
        # a string with nn_string_ix = i has its neighbors located at near_neighbors_list[i]
        # a string with near_neighbor_nn_string_ix = i is the string located at nn_strings_this_index[i]
        list_cols = ['near_neighbor_nn_string_ix', 'cos_dist']
        df = pd.DataFrame({
          col:np.repeat(df[col].values, df[list_cols[0]].str.len())
          for col in df.columns.drop(list_cols)}
        ).assign(**{list_cols[0]:np.concatenate(df[list_cols[0]].values),
                    list_cols[1]:np.concatenate(df[list_cols[1]].values)})[df.columns]

        # above, nn_string_ix is indices
        # match these indices to the string they correspond to in nn_strings_queried_this_batch
        # (and its commonness_penalty)
        nn_string_info_mini = nn_string_info_queried_this_batch[['nn_string', 'commonness_penalty']].copy()
        nn_string_info_mini.columns = ['nn_string_1', 'commonness_penalty_1']
        df = pd.merge(df, nn_string_info_mini, left_on='nn_string_ix', right_index=True)

        # above, near_neighbor_nn_string_ix is indices
        # match these indices to the string they correspond to in nn_strings_this_index
        # (and its commonness_penalty)
        nn_string_info_mini = nn_string_info_this_index[['nn_string', 'commonness_penalty']].copy()
        nn_string_info_mini.columns = ['nn_string_2', 'commonness_penalty_2']
        df = pd.merge(df, nn_string_info_mini, left_on='near_neighbor_nn_string_ix', right_index=True)

        df = df[['nn_string_ix', 'nn_string_1', 'nn_string_2', 'cos_dist', 'commonness_penalty_1', 'commonness_penalty_2']]

        return df

    def get_exact_match_candidate_pairs(self, nn_string_info_multi, nn_string_expanded_df, blocking_thresholds):
        '''All nn_strings that appear more than once need to have a corresponding
        nn_string, nn_string candidate pair -- we can skip the "approximation" easily
        for this type of candidate pair.

        Args:
            nn_string_info_multi (pd.DataFrame): nn_string_info, subset to nn_strings with n_new > 0 & n_total > 1
            nn_string_expanded_df (pd.DataFrame): table at nn_string/ed_string/absval_string level (expanded if split_name is True)
            blocking_thresholds (dict):  dictionary with thresholds for blocking, e.g. high and low bar

        Returns:
            pd.DataFrame: portion of the candidate pairs list (where nn_string_1 == nn_string_2)

            ==================   ==============================================================
            nn_string            concatenated version of nn-blocking columns (sep by ::)
            commonness_penalty   float indicating how common the last name is
            n_new                number of times this nn_string appears in a "new" record
            n_existing           number of times this nn_string appears in an "existing" record
            n_total              number of times this nn_string appears in any record
            ==================   ==============================================================
        '''

        logger.info('Getting identical candidate pairs.')

        # adding (nnsX, nnsX) as a candidate if there are 2+ record for bsX
        # and 1+ come from "new"; second part of condition is already met
        # given that this df has just "to query" names)

        nn_string_info_multi = nn_string_info_multi.copy()

        # identical first and last names
        exact_match_df = pd.DataFrame(data={
            'nn_string_1' : nn_string_info_multi.nn_string,
            'nn_string_2' : nn_string_info_multi.nn_string,
            'cos_dist' : 0,
            'commonness_penalty_1' : nn_string_info_multi.commonness_penalty,
            'commonness_penalty_2' : nn_string_info_multi.commonness_penalty
        })

        exact_match_cp_df = self.apply_blocking_filter(exact_match_df, blocking_thresholds,
                nn_string_expanded_df, nns_match=True)

        return exact_match_cp_df


def get_blocking_columns(blocking_scheme):
    '''Get the list of blocking variables for each type of blocking:

    Args:
        blocking_scheme (dict): dictionary with info on how to do blocking

    Returns:
        list of string list: the variable names needed for each type of blocking
    '''

    nn_cols = blocking_scheme['cosine_distance']['variables']
    ed_col = blocking_scheme['edit_distance']['variable']

    try:
        absval_col = blocking_scheme['absvalue_distance']['variable']
        if absval_col is None:
            raise
        cols_to_read = nn_cols + [ed_col, absval_col, 'file_type', 'drop_from_nm', 'record_id']
        logger.trace('Absvalue filter used as backup to edit_distance filter.')
    except:
        absval_col = None
        cols_to_read = nn_cols + [ed_col, 'file_type', 'drop_from_nm', 'record_id']
        logger.trace('No backup to edit_distance filter being used.')

    return nn_cols, ed_col, absval_col


def read_an(an_file, nn_cols, ed_col, absval_col):
    '''Read in relevant columns for blocking from the all-names file.

    Args:
        an_file (str): path to the all-names file
        nn_cols (list of strings): variables for near neighbor blocking
        ed_col (list of strings): variables for edit-distance blocking
        absval_col (list of strings): variables for absolute-value blocking

    Returns:
        pd.DataFrame: all-names dataframe, relevant columns only (where drop_from_nm == 0)

        =======================   =======================================================
        record_id                 unique record identifier
        blockstring               concatenation of all the blocking variables (sep by ::)
        file_type                 either "new" or "existing"
        drop_from_nm              flag, 1 if met any "to drop" criteria 0 otherwise
        <nn-blocking column(s)>   variables for near-neighbor blocking
        <ed-blocking column>      variable for edit-distance blocking
        <av-blocking column>      (optional) variable for abs-value blocking
        nn_string                 concatenated version of nm-blocking columns (sep by ::)
        ed_string                 copy of ed-blocking column
        absval_string             copy of ed-blocking column
        =======================   =======================================================
    '''

    cols_to_read = nn_cols + [ed_col, \
            'blockstring', 'file_type', 'drop_from_nm', 'record_id']

    if absval_col is not None:
        cols_to_read = nn_cols + [ed_col, absval_col, \
                'blockstring', 'file_type', 'drop_from_nm', 'record_id']

    table = pq.read_table(an_file)
    an =  table.to_pandas()
    if absval_col is not None and absval_col not in an.columns.tolist():
        logger.error(f'The {absval_col} variable was not found, and is needed for the chosen blocking scheme.')
        raise ValueError
    an = an[cols_to_read]

    an = an.fillna('')
    an = an[an.drop_from_nm == 0]

    an['nn_string'] = np.vectorize(get_nn_string_from_blockstring,
                                   otypes=['object'])(an.blockstring)
    an['ed_string'] = an[ed_col]
    if absval_col is not None:
        an['absval_string'] = an[absval_col]
        an.loc[an.absval_string == '', 'absval_string'] = np.NaN
        an['absval_string'] = an.absval_string.astype(float)

    return an


def get_nn_string_counts(an):
    '''Count number of records per nn_strings (per file_type).

    Args:
        an (pd.DataFrame): all-names table, relevant columns only (where drop_from_nm == 0)

            =======================  =======================================================
            record_id                unique record identifier
            blockstring              concatenation of all the blocking variables (sep by ::)
            file_type                either "new" or "existing"
            drop_from_nm             flag, 1 if met any "to drop" criteria 0 otherwise
            <nn-blocking column(s)>  variables for near-neighbor blocking
            <ed-blocking column>     variable for edit-distance blocking
            <av-blocking column>     (optional) variable for abs-value blocking
            nn_string                concatenated version of nm-blocking columns (sep by ::)
            ed_string                copy of ed-blocking column
            absval_string            copy of ed-blocking column
            =======================  =======================================================

    Return:
        dict: two keys (new and existing), mapping to a dictionary of nn_strings to n_records
    '''

    an = an.copy()

    logger.trace('Counting nn_strings.')
    temp = an.groupby(['file_type', 'nn_string']).size().reset_index().pivot(
            index='nn_string', columns='file_type')
    temp.columns = temp.columns.droplevel()
    temp = temp.fillna(0)
    nn_strings_to_num_recs = temp.to_dict('dict')

    if 'existing' not in nn_strings_to_num_recs:
        nn_strings_to_num_recs['existing'] = defaultdict(int)

    return nn_strings_to_num_recs


def get_common_name_penalties(clean_last_names, max_penalty, num_threshold_bins=1000):
    '''Create a dictionary mapping each last name to a "commonness penalty." Two SMITHs are
    less likely to be the same person than two HANDAs, since SMITH is such a common name. This
    function quantifies this penalty for use in later blocking calculations. A more common name
    recieves a higher number, topping out at max_penalty.

    Args:
        clean_last_names (pd.Series): clean (un-split) last name column (one row per record)
        max_penalty (float): the maximum penalty (for the most common names)
        num_threshold_bins (int): number of different categories of commonnness to create

    Returns:
        dict: dictionary mapping name (str) to penalty (float)
    '''

    logger.trace('Getting last names frequencies.')
    last_name_freq_df = clean_last_names.groupby(clean_last_names).size()
    last_name_freq_df.index.name = 'last_name'
    last_name_freq_df.name = 'ln_count'
    last_name_freq_df = last_name_freq_df.reset_index()

    commonness_penalty_lookup = defaultdict(lambda: 0)

    last_name_freq_df['norm_ln_count'] = \
            last_name_freq_df.ln_count / last_name_freq_df.ln_count.sum()
    last_name_freq_df = last_name_freq_df.sort_values('norm_ln_count', ascending=False)

    last_name_freq_df['bin'] = pd.qcut(
            last_name_freq_df.norm_ln_count,
            num_threshold_bins,
            labels=False, duplicates='drop')
    threshold_bins = last_name_freq_df['bin'].unique()
    threshold_levels = pd.DataFrame(data={
            'bin' : threshold_bins,
            'threshold' : np.linspace(max_penalty, 0, len(threshold_bins))})
    last_name_freq_df = pd.merge(last_name_freq_df, threshold_levels, on='bin')
    last_name_freq_df.set_index('last_name', inplace=True)
    commonness_penalty_lookup = defaultdict(
            lambda: max_penalty, last_name_freq_df.threshold.to_dict())

    return commonness_penalty_lookup


def get_all_shingles():
    '''Get all valid 2-shingles.

    Returns:
        list: valid 2-shingles
    '''

    valid_characters = string.ascii_uppercase + ' *'
    all_tuples = list(itertools.product(valid_characters,valid_characters))
    all_two_shingles = [tup[0] + tup[1] for tup in all_tuples]
    valid_two_shingles = [sh for sh in all_two_shingles if sh not in ['  ', '* ', ' *']]
    # NOTE: leave ** in case we ever allow blank nn_strings

    return valid_two_shingles


def prep_index():
    '''Initialize index data structure, which will store similarity information
    about the names, and load processed shingles into it.

    Returns:
        nnmslib.FloatIndex: nmslib index object (pre time-consuming build call)
    '''

    # intialize index
    space_type = 'cosinesimil_sparse'
    space_params = {}
    method_name = 'hnsw'

    ix = nmslib.init(
        space=space_type,
        space_params=space_params,
        method=method_name,
        data_type=nmslib.DataType.SPARSE_VECTOR,
        dtype=nmslib.DistType.FLOAT)

    return ix


def get_second_index_nn_strings(all_nn_strings, main_nn_strings):
    '''Get nn_strings that haven't already been stored in the main index.

    Args:
        all_nn_strings (list): list of all nn_strings in the data (expanded if split_names is True)
        main_nn_strings (list): list of nn_strings already in the main index

    Returns:
        list: the nn_strings that are not in main_nn_strings
    '''

    all_nn_strings = all_nn_strings[:]
    main_nn_strings = main_nn_strings[:]

    all_nn_strings_s = pd.Series(all_nn_strings)
    second_index_nn_strings = all_nn_strings_s[
            all_nn_strings_s.isin(set(main_nn_strings)) == False].tolist()

    return second_index_nn_strings


def save_main_index(main_index, main_index_nn_strings, main_index_file):
    '''Save the main nmslib index and pickle dump the associated nn_strings list.

    Args:
        main_index (nmslib.FloatIndex): the main, built nmslib index
        main_index_nn_strings (list): list of nn_strings in the main index
        main_index_file (str): path to store the main nmslib index
    '''

    main_index.saveIndex(main_index_file, save_data=True)

    with open(main_index_file + '.pkl', 'wb') as pf:
        pickle.dump(main_index_nn_strings, pf, protocol=pickle.HIGHEST_PROTOCOL)


def load_main_index_nn_strings(og_blocking_index_file):
    '''Load the nn_strings that are in an existing nmslbi index file.

    Args:
        og_blocking_index_file (str): path to original blocking index

    Returns:
        list: loaded list of nn_strings in an existing nmslib index
    '''

    with open(og_blocking_index_file + '.pkl', 'rb') as f:
        main_index_nn_strings = pickle.load(f)

    return main_index_nn_strings


def write_some_cps(cand_pairs, candidate_pairs_file, header=False):
    '''Write out a portion of the candidate-pairs.

    Args:
        cand_pairs (pd.DataFrame): chunk of the candidate-pairs file
        candidate_pairs_file (str): path to the candidate-pairs file
        header (bool): True if this is the first time calling this function
    '''

    cand_pairs.to_csv(
            candidate_pairs_file,
            mode='a',
            index=False,
            header=header,
            quoting=csv.QUOTE_NONNUMERIC)


def generate_true_pairs(must_links_df):
    """Reduce the must-link records pairs must-link blockstring pairs.

    Args:
        must_links_df (pd.DataFrame): list of must-link record pairs

            ===================   =======================================================
            record_id_1           unique identifier for the first record in the pair
            record_id_2           unique identifier for the second record in the pair
            blockstring_1         blockstring for the first record in the pair
            blockstring_2         blockstring for the second record in the pair
            drop_from_nm_1        flag, 1 if the first record in the pair was not eligible for matching
            drop_from_nm_2        flag, 1 if the second record in the pair was not eligible for matching
            existing              flag, 1 if the pair is must-link because of ExistingID
            ===================   =======================================================

    Return:
        pd.DataFrame: list of must-link blockstring pairs (where both record have drop_from_nm == 0)

        ===================   =======================================================
        blockstring_1         blockstring for the first record in the pair
        blockstring_2         blockstring for the second record in the pair
        ===================   =======================================================
    """

    must_links_df = must_links_df[(must_links_df.drop_from_nm_1 == 0) &
                                  (must_links_df.drop_from_nm_2 == 0)]

    true_pairs_df = must_links_df[['blockstring_1', 'blockstring_2']].drop_duplicates()

    return true_pairs_df




import pandas as pd
import numpy as np
import tempfile

from namematch.model_evaluation_functions import get_accuracy


def test_get_accuracy():

    thresh = 0.5

    df = pd.DataFrame({
        'phat':[0, .4, .5, .6, .9],
        'outcome':[0, 0, 1, 1, 1],
        'weight':[1, 1, 1, 1, 1]
    })

    assert get_accuracy(df, thresh) == 1

    one_wrong_no_weights_df = pd.DataFrame({
        'phat':[0, .9, .5, .6, .9],
        'outcome':[0, 0, 1, 1, 1],
        'weight':[1, 1, 1, 1, 1]
    })
    
    one_wrong_weights_df = pd.DataFrame({
        'phat':[0, .9, .5, .6, .9],
        'outcome':[0, 0, 1, 1, 1],
        'weight':[1, 2, 1, 1, 1]
    })

    assert get_accuracy(one_wrong_no_weights_df, thresh) > get_accuracy(one_wrong_weights_df, thresh)




import os

from functools import wraps, partial

from line_profiler import LineProfiler


class Profiler:

    line_profiler = LineProfiler()

    def __call__(self, func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            if args[0].enable_lprof:
                self.line_profiler.add_function(func)
                self.line_profiler.enable_by_count()
            return func(*args, **kwargs)
        return wrapper




from setuptools import find_packages, setup
from pathlib import Path

ROOT_PATH = Path(__file__).parent

README_PATH = ROOT_PATH / "README.md"

LICENSE_PATH = ROOT_PATH / "LICENSE"

with open('README.md') as readme_file:
    readme = readme_file.read()

def get_requirements(requirements_file='requirements.txt'):
    """Get the contents of a file listing the requirements"""
    lines = open(requirements_file).readlines()
    dependencies = []
    for line in lines:
        dep = line.strip()
        if dep.startswith('git+'):
            # VCS reference for dev purposes, expect a trailing comment
            # with the normal requirement
            __, __, repo = dep.rpartition('=')
            dep = repo + ' @ ' + dep
        if dep:
            dependencies.append(dep)
    return dependencies

setup(
    name='namematch',
    version='1.2.1',
    description='Tool for probabilistically linking the records of individual entities (e.g. people) within and across datasets',
    long_description=README_PATH.read_text(),
    long_description_content_type="text/markdown",
    author="University of Chicago Crime Lab",
    author_email='mmcneill@uchicago.edu, tlin2@uchicago.edu, zjelveh@umd.edu',
    url='https://github.com/urban-labs/namematch',
    project_urls={
        "Documentation": "https://urban-labs.github.io/namematch/",
        "Source Code": "https://github.com/urban-labs/namematch",
    },
    packages=find_packages(),
    include_package_data=True,
    install_requires=get_requirements('requirement/main.txt'),
    package_data={'': ['*.yaml']},
    entry_points={
        'console_scripts': [
            'namematch = namematch.cli:execute']},
    license='AGPL-3.0',
    keywords=['record linkage'],
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ],
    python_require='>=3.7',
    test_suite='tests',
    tests_require=get_requirements('requirement/dev.txt')
)




from unittest.mock import patch

import pandas as pd
import numpy as np
from namematch.cluster import Cluster, Constraints


def test_get_initial_clusters(params_and_schema, all_names_clustering_df, must_links_df, logger_for_testing):
    # test from scratch
    params, schema = params_and_schema
    with patch('namematch.cluster.logger', logger_for_testing) as mock_debug:
        cluster = Cluster(
                params,
                schema,
                constraints=None,
                must_links_file=None,
                potential_edges_dir=None,
                flipped0_edges_file=None,
                all_names_file=None,
                cluster_assignments=None,
        )
        eid_col = None

        clusters, cluster_assignments, original_cluster_ids = \
                cluster.get_initial_clusters(must_links_df, all_names_clustering_df, eid_col)

        assert original_cluster_ids is None
        assert 3 == len(clusters)
        assert {"A", "B", "C", "D", "E", "F", "G", "H"} == set(cluster_assignments.keys())

        # test incremental

        # TODO


def test_auto_is_valid_edge(params_and_schema):

    edges_df = pd.DataFrame({
        'ir_1':['12345', np.NaN, '45678', '343252'],
        'ir_2':['12346', np.NaN, '45678', '982834']
    })
    params, schema = params_and_schema
    auto_valid_potential_edges = Cluster(params, schema).auto_is_valid_edge(
        edges_df.copy(), uid_cols=['ir'], allow_clusters_w_multiple_unique_ids=False,
        leven_thresh=None, eid_col=None)
    assert len(auto_valid_potential_edges) == 2

    auto_valid_potential_edges = Cluster(params, schema).auto_is_valid_edge(
        edges_df.copy(), uid_cols=['ir'], allow_clusters_w_multiple_unique_ids=False,
        leven_thresh=1, eid_col=None)
    assert len(auto_valid_potential_edges) == 3

    edges_df = pd.DataFrame({
        'ir_1':['12345', 'melissa', np.NaN, '45678', '343252', '17'],
        'ir_2':['12346', 'merissa', np.NaN, '45678', '982834', '17'],
        'ssn_1':['333', 'abc', 'mcne ill', '999', np.NaN, np.NaN],
        'ssn_2':['333', 'abc', 'mcneill', np.NaN, np.NaN, np.NaN],
    })
    auto_valid_potential_edges = Cluster(params, schema).auto_is_valid_edge(
        edges_df.copy(), uid_cols=['ir', 'ssn'], allow_clusters_w_multiple_unique_ids=False,
        leven_thresh=None, eid_col=None)
    assert len(auto_valid_potential_edges) == 4

    auto_valid_potential_edges = Cluster(params, schema).auto_is_valid_edge(
        edges_df.copy(), uid_cols=['ir', 'ssn'], allow_clusters_w_multiple_unique_ids=False,
        leven_thresh=1, eid_col=None)
    assert len(auto_valid_potential_edges) == 5


def test_auto_is_valid_cluster():
    pass


def test_get_potential_edges():
    pass


def test_load_cluster_info(all_names_parquet_file, params_and_schema):

    params, schema = params_and_schema

    def default_get_cols_used():
        return 'all'

    def no_get_cols_used():
        return {}

    def some_get_cols_used():
        return {'dob':'date'}

    default = Constraints()
    default.get_columns_used = default_get_cols_used

    no_cols = Constraints()
    no_cols.get_columns_used = no_get_cols_used

    some_cols = Constraints()
    some_cols.get_columns_used = some_get_cols_used

    # test default get_columns_used() function
    cluster_info = Cluster(params, schema).load_cluster_info(
        all_names_parquet_file, uid_cols=['uid'], eid_col=None, cluster_logic=default)
    assert cluster_info.shape[1] == 12
    assert (cluster_info.dtypes == 'object').all()

    # test empty get_columns_used() function
    cluster_info = Cluster(params, schema).load_cluster_info(
        all_names_parquet_file, uid_cols=['uid'], eid_col=None, cluster_logic=no_cols)
    assert cluster_info.shape[1] == 2
    assert (cluster_info.dtypes == 'object').all()

    # test standard get_columns_used() function
    cluster_info = Cluster(params, schema).load_cluster_info(
        all_names_parquet_file, uid_cols=['uid'], eid_col=None, cluster_logic=some_cols)
    assert cluster_info.shape[1] == 3
    assert (cluster_info.dtypes == 'object').sum() == 2
    assert (cluster_info.dtypes == 'datetime64[ns]').sum() == 1


def test_cluster_potential_edges():
    pass


def test_pandas_sorting_default():

    df = pd.DataFrame(data={
        'a':[1, 2, 3, 4, 5],
        'b':[1, .99, .98, .97, .8]
    })
    df_dict_list = df.to_dict('records')

    a_vec = [d['a'] for d in df_dict_list]
    b_vec = [d['b'] for d in df_dict_list]

    assert df.a.tolist() == a_vec
    assert df.b.tolist() == b_vec





import pandas as pd
from unittest.mock import patch

from namematch.data_structures.variable import Variable
from namematch.process_input_data import (
    ProcessInputData,
    process_set_missing,
    process_drop
)


def test_process_geo_column(logger_for_testing, raw_data_df, params_and_schema):
    # load fake data
    v = Variable({"name": "lat/lon"})
    raw_data_df["lat"] = raw_data_df["lat"].astype(str)
    raw_data_df["lon"] = raw_data_df["lon"].astype(str)

    with patch('namematch.process_input_data.logger', logger_for_testing) as mock_debug:
        # test two cols
        params, schema = params_and_schema
        processed_geo = ProcessInputData(params, schema).process_geo_column(raw_data_df[["lat", "lon"]], v)
        assert 1 == processed_geo.shape[1]
        assert raw_data_df.shape[0] == processed_geo.shape[0]

        # test one col
        processed_geo = ProcessInputData(params, schema).process_geo_column(raw_data_df[["lat/lon"]], v)
        assert 1 == processed_geo.shape[1]
        assert raw_data_df.shape[0] == processed_geo.shape[0]


def test_parse_address(params_and_schema):
    test_cases = [
        ("123 Main St.", "123", "main", "street"),
        ("123 Broadway", "123", "broadway", ""),
        ("", "", "", ""),
        ("123 MAIN STREET", "123", "main", "street"),
        ("123 MAIN ST", "123", "main", "street"),
        # ("666 Lincoln Terrace", "666", "lincoln", "terrace"), see TODO in generate_data_files
        ("Twelve Forbes Ave", "12", "forbes", "avenue")]
    params, schema = params_and_schema
    process_input_data = ProcessInputData(params, schema)
    for (address, expected_num, expected_name, expected_type) in test_cases:
        parsed_address = process_input_data.parse_address(address)
        assert len(parsed_address) == 3
        (num, name, typ) = parsed_address
        assert expected_num == num
        assert expected_name == name
        assert expected_type == typ


def test_process_address_column(raw_data_df, params_and_schema):
    params, schema = params_and_schema
    # process
    processed_address = ProcessInputData(params, schema).process_address_column(raw_data_df[["address"]])

    # test
    assert raw_data_df.shape[0] == processed_address.shape[0]
    assert 3 == processed_address.shape[1]


def test_process_check_name_items(logger_for_testing, params_and_schema):
    with patch('namematch.process_input_data.logger', logger_for_testing) as mock_debug:
        # test that the series name gets set and
        # items are stripped and upper case
        name = "test_name"
        v = Variable({"name": name, "compare_type": "", "check": ""})
        s = pd.Series([" test1", "test2", " test3 "])
        params, schema = params_and_schema
        processed_s = ProcessInputData(params, schema).process_check(s, v)
        expected_s = pd.Series(["TEST1", "TEST2", "TEST3"])
        expected_s.name = name
        pd.testing.assert_series_equal(expected_s, processed_s)


def test_process_check_numeric_strings(logger_for_testing, params_and_schema):
    with patch('namematch.process_input_data.logger', logger_for_testing) as mock_debug:
        # test that numeric strings are checked
        name = "test_name"
        v = Variable({"name": name, "compare_type": "", "check": "Numeric"})
        s = pd.Series(["123", "5555", "1997", "notnumeric", "alsonot"])
        params, schema = params_and_schema
        processed_s = ProcessInputData(params, schema).process_check(s, v)
        expected_s = pd.Series(["123", "5555", "1997", "", ""])
        expected_s.name = name
        pd.testing.assert_series_equal(expected_s, processed_s)


def test_process_check_dates(logger_for_testing, params_and_schema):
    with patch('namematch.process_input_data.logger', logger_for_testing) as mock_debug:
        # test that dates are checked
        name = "test_name"
        v = Variable({"name": name, "compare_type": "", "check": "Date - %y-%m-%d"})
        s = pd.Series(["97-6-5", "85-10-2", "", "0000"])
        params, schema = params_and_schema
        processed_s = ProcessInputData(params, schema).process_check(s, v)
        expected_s = pd.Series(["1997-06-05", "1985-10-02", '', ''])
        expected_s.name = name
        pd.testing.assert_series_equal(expected_s, processed_s)


def test_process_check_categorical_variables(logger_for_testing, params_and_schema):
    with patch('namematch.process_input_data.logger', logger_for_testing) as mock_debug:
        # test that categorical variables are checked
        name = "test_name"
        v = Variable({"name": name, "compare_type": "Categorical", "check": "f,m,other"})
        s = pd.Series(["m", "f", "other", "", "mm", "m", "mf", "other"])
        params, schema = params_and_schema
        processed_s = ProcessInputData(params, schema).process_check(s, v)
        expected_s = pd.Series(["m", "f", "other", "", "", "m", "", "other"])
        expected_s.name = name
        pd.testing.assert_series_equal(expected_s, processed_s)

def test_process_set_missing(logger_for_testing):
    with patch('namematch.process_input_data.logger', logger_for_testing) as mock_debug:
        input_s = pd.Series(["hello", "test1", "NA", "0000", "NA", "test2", "0000"])
        expected_s = pd.Series(["hello", "test1", ""  ,  ""   , ""  , "test2", ""])
        processed_s = process_set_missing(input_s, ["NA", "0000"])
        pd.testing.assert_series_equal(expected_s, processed_s)

def test_process_drop(logger_for_testing):
    with patch('namematch.process_input_data.logger', logger_for_testing) as mock_debug:
        s = pd.Series(["hello", "test1", "NA", "0000", "NA", "test2", "0000"])
        expected_drops = [2, 3, 4, 6]
        processed_drops = process_drop(s, ["NA", "0000"])
        assert expected_drops == processed_drops


def test_process_auto_drops():
    pass


def test_process_data():
    pass


def test_split_last_names():
    pass





import yaml
import logging
from copy import deepcopy

from namematch.utils.utils import dump_yaml, to_dict, load_yaml

logger = logging.getLogger()

params_lookup = {
    'process_input_data': [
        'auto_drop_logic',
        'blocking_scheme',
        'last_name_column',
        'verbose',
        'incremental',
        'input_data_batch_size'
        ],
    'generate_must_links': [],
    'block': [
        'split_names',
        'last_name_column',
        'blocking_scheme',
        'blocking_thresholds',
        'incremental',
        'index',
        'num_workers',
        'nmslib',
        'verbose',
        'parallelize'
        ],
    'generate_data_rows': [
        'verbose',
        'data_rows_batch_size',
        'num_workers',
        'match_train_criteria',
        'first_name_column',
        'last_name_column',
        'parallelize',
        'incremental',
        'exact_match_variables',
        'leven_thresh',
        ],
    'fit_model': [
        'max_selection_train_eval_n',
        'missingness_model',
        'weight_using_selection_model',
        'default_threshold',
        'optimize_threshold',
        'fscore_beta',
        'match_train_criteria',
        'pct_train',
        'max_match_train_n',
        'allow_clusters_w_multiple_unique_ids',
        ],
    'predict': [
        'use_uncovered_phats',
        'parallelize',
        ],
    'cluster': [
        'initialize_from_ground_truth_1s',
        'incremental',
        'verbose',
        'allow_clusters_w_multiple_unique_ids',
        'leven_thresh',
        ],
    'generate_output': [
        'blocking_scheme',
        ],
}

class Parameters():
    '''Class that houses important matching parameters. Handles validation of the config file.'''

    def __init__(self, validate_param_dict):

        for param, param_value in validate_param_dict.items():
            setattr(self, param, param_value)

    @staticmethod
    def check_integrity(defaults, param, param_value):
        '''Ensure that parameters are of the appropriate type.

        Args:
            param (str): parameter name (key)
            param_value: value of the given key  (parameter name)
        '''

        params__start_with_alpha = []
        params__numeric = []
        params__positive_numeric = [
                'fscore_beta', 'leven_thresh', 'num_workers', 'pct_train',
                'secondary_index_limit', 'verbose', 'input_data_batch_size', 'data_rows_batch_size']
        params__boolean = [
                'use_uncovered_phats', 'allow_clusters_w_multiple_unique_ids']
        params__specific_value = {
                'rebuild_main_index' : ['always', 'never', 'if_secondary_index_exceeds_limit']}

        valid = True

        same_as_default = (param_value == defaults[param])
        if defaults[param] is None and param_value is None:
            same_as_default = True

        if not same_as_default:
            if param in params__start_with_alpha:
                if not str.isalpha(param_value[0]):
                    valid = False
            if param in params__numeric:
                if not isinstance(param_value, (int, float, complex)):
                    valid = False
            if param in params__positive_numeric:
                if not isinstance(param_value, (int, float, complex)) or param_value <= 0:
                    valid = False
            if param in params__boolean:
                if not isinstance(param_value, bool):
                    valid = False
            if param in params__specific_value:
                if param_value not in params__specific_value[param]:
                    valid = False

        if not valid:
            logger.warning(f"The {param} parameter has taken an illegal value "
                           f"({param_value}). Default value ({defaults[param]}) "
                           f"will be used.")

        return valid

    @classmethod
    def init(cls, config: dict, defaults: dict):
        '''Create a Parameters instance.

        Args:
            config (dict): dictionary with match parameter values
            defaults (dict): dictionary with default params

        Returns:
            :mod:`namematch.data_structures.parameters.Parameters`: instance of the Parameters class
        '''

        param_dict = {}

        # issue warning if user specifies param not understood by code
        params_default = list(defaults.keys())[:]
        params_user_specified = list(config.keys())
        params_acceptable_non_default = ['existing_data_files',
                                         'data_files', 'variables']
        params_acceptable = params_default + params_acceptable_non_default
        params_unrecognized = [
                param for param in params_user_specified
                if param not in params_acceptable]
        if len(params_unrecognized) > 0:
            params_unrecognized = ', '.join(params_unrecognized)
            logger.warning(f"The following parameters are not "
                           f"recognized (will be ignored): {params_unrecognized}")
        params_only_incremental = []
        for param in params_only_incremental:
            if 'existing_data_files' not in config and param in config:
                logger.warning(f"The following parameters are not recognized "
                               f"for non-incremental matching (will be "
                               f"ignored): {param}")

        # add private params based on input params
        param_dict.update(defaults)

        # overwrite defaults and private config values with values defined in config
        for param, param_value in config.items():
            if isinstance(param_value, list) and param != 'variables':
                for value in param_value:
                    if (not Parameters.check_integrity(defaults, param, value)) and (param in params_default):
                        raise Exception(f"parameter '{param}' contains value '{value}' which is an invalid type")
                param_dict[param] = param_value
            else:
                if param in params_default:
                    if Parameters.check_integrity(defaults, param, param_value):
                        param_dict[param] = param_value
                    param_dict[param] = param_value
        param_dict['incremental'] = (len(config.get('existing_data_files', {})) > 0)

        if param_dict['num_workers'] > 1:
            param_dict['parallelize'] = True

        return cls(param_dict)

    @classmethod
    def load(cls, filepath):
        '''Load a Parameters instance.

        Args:
            filepath (str): path to a yaml version of a Parameters instance

        Returns:
            :mod:`namematch.data_structures.parameters.Parameters`: instance of the Parameters class
        '''

        param_dict = load_yaml(filepath)

        # NOTE: no need for validation because simple loading
        # an object that was validated when initialized

        return cls(param_dict)

    @classmethod
    def load_from_dict(cls, param_dict):
        return cls(param_dict)

    def check_for_required_variables(self, variables):
        '''Validate that the config includes required variables.'''

        missing_required_variables = []
        for required_var in self.required_variables:
            if required_var not in variables.get_names():
                missing_required_variables.append(required_var)

        if len(missing_required_variables) > 0:
            error_str = ', '.join(missing_required_variables)
            logger.error(f'The following required variable names '
                         f'are missing: {error_str}')
            raise ValueError

    def validate_exactmatch_variables(self, variables):
        '''Validate that the exact_match_variables and negate_exact_match_variables parameters.'''

        em_vars = self.exact_match_variables + self.negate_exact_match_variables

        invalid_variables = []
        valid_variables = variables.get_names()
        for variable in em_vars:
            if variable not in valid_variables:
                invalid_variables.append(variable)

        if len(invalid_variables) > 0:
            invalid_variables = ', '.join(invalid_variables)
            logger.warning("The exactmatch flag could not be calculated because the following "
                           f"variables are not defined: {invalid_variables}. You may need to adjust "
                           "the exact_match_variables and negate_exact_match_variables parameters.")


    def validate_blocking_scheme(self, variables):
        '''Validate that the blocking scheme is in the correct format and provieds the minimum
        number of blocking variables per blocking type (cosine_distance, edit_distance, absvalue_distance).'''

        blocking_params = ['cosine_distance', 'edit_distance',
                           'absvalue_distance', 'alpha', 'power']
        if any([v not in blocking_params for v in list(self.blocking_scheme.keys())]):
            logger.error("Blocking scheme is misspecified.")
            raise ValueError

        if len(self.blocking_scheme['cosine_distance']['variables']) > 2:
            logger.warning(f"Only the first two variables in blocking_scheme"
                           f"['cosine_distance'] will be used.")
            self.blocking_scheme['cosine_distance']['variable'] = \
                    self.blocking_scheme['cosine_distance']['variable'][0:2]

        if type(self.blocking_scheme['edit_distance']['variable']) == list:
            logger.warning(f"Only one variable allowed in blocking_scheme"
                           f"['edit_distance']; using first one.")
            self.blocking_scheme['edit_distance']['variable'] = \
                    self.blocking_scheme['edit_distance']['variable'][0]

        alpha = self.blocking_scheme['alpha']
        power = self.blocking_scheme['power']
        if not isinstance(alpha, (int, float)) or not isinstance(power, float):
            logger.error(f"Incorrect parameter definitions in blocking_scheme: "
                         f"alpha (numeric) and power (float).")
            raise ValueError

        invalid_variables = []
        valid_variables = variables.get_names()
        blocking_variables = self.get_blocking_variables()
        for variable in blocking_variables:
            if variable not in valid_variables:
                invalid_variables.append(variable)

        if len(invalid_variables) > 0:
            invalid_variables = ', '.join(invalid_variables)
            logger.error(f'The following variable names, needed for the specified '
                         f'blocking scheme, are missing: {invalid_variables}')
            raise ValueError

    def get_blocking_variables(self):
        '''Get list of blocking variable nicknames.

        Return:
            list of variable nicknames (all-names columns) to use for blocking
        '''

        bv = self.blocking_scheme['cosine_distance']['variables'][:]
        bv.append(self.blocking_scheme['edit_distance']['variable'])
        #bv.append(self.blocking_scheme['absvalue_distance']['variable'])
        bv = [bv for bv in bv if bv is not None]

        return bv

    def validate(self, variables):
        '''Validate several components of the config file.'''

        self.check_for_required_variables(variables)
        self.validate_blocking_scheme(variables)
        self.validate_exactmatch_variables(variables)

    def write(self, output_file):
        '''Write the Parameters to a yaml file.

        Args:
            output_file (str): path to write parameter dictionary
        '''

        dump_yaml(to_dict(self), output_file)

    def copy(self):
        '''Create a deep copy of a Parameters object.'''
        return deepcopy(self)

    def stage_params_lookup(self):
        stage_params_lookup = {}
        for k, v in params_lookup.items():
            stage_params_lookup[k] = {p: getattr(self, p, None) for p in v}
        return stage_params_lookup

    def get_stage_params(self, stage):
        return self.stage_params_lookup()[stage]



import editdistance
import jellyfish
import logging
import numpy as np
import pandas as pd

from datetime import datetime
from math import sqrt
from pyjarowinkler import distance

logger = logging.getLogger()

try:
    profile
except:
    from line_profiler import LineProfiler
    profile = LineProfiler()


@profile
def get_name_probabilities(df, np_object, first_name_col, last_name_col):
    '''Use the name-probabilities package to assesess the likelihood of a pair of names
    being the same person.

    Args:
        df (pd.DataFrame): side-by-side table
        np_object (NameMatcher object): contains info about name probabilities
        first_name_col (str): name of first-name column
        last_name_col (str): name of last-name column

    Returns:
        pd.DataFrame: side-by-side table with a few new name-probability columns
    '''

    if np_object is None:
        return df

    df = df.copy()

    df['fn1'] = df[first_name_col + '_1'].str.replace(' ', '')
    df['fn2'] = df[first_name_col + '_2'].str.replace(' ', '')
    df['ln1'] = df[last_name_col + '_1'].str.replace(' ', '')
    df['ln2'] = df[last_name_col + '_2'].str.replace(' ', '')
    df['fn_ln_1'] = '*' + df.fn1 + ' ' + df.ln1 + '*'
    df['fn_ln_2'] = '*' + df.fn2 + ' ' + df.ln2 + '*'
    df['rev_fn_ln_1'] = '*' + df.ln1 + ' ' + df.fn1 + '*'
    df['rev_fn_ln_2'] = '*' + df.ln2 + ' ' + df.fn2 + '*'

    df['prob_name_1'] = df.fn_ln_1.apply(lambda x: np_object.probName(x))
    df['prob_name_2'] = df.fn_ln_2.apply(lambda x: np_object.probName(x))

    df['prob_rev_name_1'] = df.rev_fn_ln_1.apply(lambda x: np_object.probName(x))
    df['prob_rev_name_2'] = df.rev_fn_ln_2.apply(lambda x: np_object.probName(x))

    df['count_pctl_name_1'] = df.fn_ln_1.map(np_object.n_name_appearances_dict)
    df['count_pctl_name_2'] = df.fn_ln_2.map(np_object.n_name_appearances_dict)

    df['count_pctl_fn_1'] = df.fn1.map(np_object.n_firstname_appearances_dict)
    df['count_pctl_fn_2'] = df.fn2.map(np_object.n_firstname_appearances_dict)

    df['count_pctl_ln_1'] = df.ln1.map(np_object.n_lastname_appearances_dict)
    df['count_pctl_ln_2'] = df.ln2.map(np_object.n_lastname_appearances_dict)

    df['prob_same_name'] = np.vectorize(np_object.probSamePerson, otypes=['float'])(
            df.fn_ln_1.values, df.fn_ln_2.values)
    df['prob_same_name_rev_1'] = np.vectorize(np_object.probSamePerson, otypes=['float'])(
            df.rev_fn_ln_1.values, df.fn_ln_2.values)
    df['prob_same_name_rev_2'] = np.vectorize(np_object.probSamePerson, otypes=['float'])(
            df.fn_ln_1.values, df.rev_fn_ln_2.values)

    df.drop(columns=['fn1', 'fn2', 'ln1', 'ln2', 'fn_ln_1',
            'fn_ln_2', 'rev_fn_ln_1', 'rev_fn_ln_2'], inplace=True)

    return df


@profile
def try_switch_first_last_name(df, first_name_col, last_name_col):
    '''The input df to this function is the output of get_name_probabilities. Using the probabilities
    calculated above, it decides whether the names labelled as FIRST and LAST were swapped. If they
    are, it fixes the swap.

    Args:
        df (pd.DataFrame): side-by-side table with a few new name-probability columns
        first_name_col (str): name of first-name column
        last_name_col (str): name of last-name column

    Returns:
        pd.DataFrame: side-by-side table with a few new name-probability columns (names swapped as needed)
    '''


    df = df.copy()

    df['full_name_1'] = df[first_name_col + '_1'] + df[last_name_col + '_1']
    df['full_name_2'] = df[first_name_col + '_2'] + df[last_name_col + '_2']
    df['reversed_full_name_1'] = df[last_name_col + '_1'] + df[first_name_col + '_1']

    # compare name_1 to name_2, and the reversed name_1 to name_2, to see which are more similar
    df['name_ed'] = np.vectorize(editdistance.eval, otypes=['float'])(
            df.full_name_1.values, df.full_name_2.values)
    df['reversed_name_ed'] = np.vectorize(editdistance.eval, otypes=['float'])(
            df.reversed_full_name_1.values, df.full_name_2.values)

    df['least_likely_name'] = (df.prob_name_2 < df.prob_name_1).astype(int) + 1

    # if the reversed name has a lower edit distance, mark that it is likely switched
    df['switched_name'] = 0
    df.loc[df.reversed_name_ed < df.name_ed, 'switched_name'] = df.least_likely_name

    # we need temp columns to orchestrate the switch
    df['fn1'] = df[first_name_col + '_1']
    df['fn2'] = df[first_name_col + '_2']
    df['ln1'] = df[last_name_col + '_1']
    df['ln2'] = df[last_name_col + '_2']

    # for names marked as switched, fix the switch
    df.loc[df.switched_name == 1, 'fn1'] = df[last_name_col + '_1']
    df.loc[df.switched_name == 1, 'ln1'] = df[first_name_col + '_1']

    df.loc[df.switched_name == 2, 'fn2'] = df[last_name_col + '_2']
    df.loc[df.switched_name == 2, 'ln2'] = df[first_name_col + '_2']

    # clean up all the columns
    df.drop(
        columns=[
            first_name_col + '_1',
            first_name_col + '_2',
            last_name_col + '_1',
            last_name_col + '_2',
            'full_name_1',
            'full_name_2',
            'reversed_full_name_1',
            'name_ed',
            'reversed_name_ed',
            'least_likely_name'],
        inplace=True)
    df.rename(
        columns={
            'fn1': first_name_col + '_1',
            'fn2': first_name_col + '_2',
            'ln1': last_name_col + '_1',
            'ln2': last_name_col + '_2'},
        inplace=True)

    return df


@profile
def compare_strings(df, varname):
    '''Compute string distances between two records for a particular matching field.

    Args:
        df (pd.DataFrame): side-by-side table
        varname (str): name of the all-names column to compute distance measures for

    Returns:
        pd.DataFrame: features associated with the current variable
            ===========================   ============================================================
            <var>_missing                 flag, 1 if either record is missing the value for this field
            <var>_edit_dist               number of character edits between whole strings
            <var>_exact_match             flag, 1 if whole string matches
            <var>_exact_match_1st         flag, 1 if first charcater matches
            <var>_exact_match_1st2nd3rd   flag, 1 if first three characters match
            <var>__soundex                flag, 1 if the soundex codes match
            <var>__nysiis                 flag, 1 if the nysiis codes match
            <var>__jw_dist                jarowinkler distance (float)
            ===========================   ============================================================
    '''

    df = df.copy()
    features_df = pd.DataFrame(index=df.index)
    col1 = varname + '_1'
    col2 = varname + '_2'

    df['msng'] = ((df[col1] == '') | (df[col2] == '')).astype(int)
    df['valid_nysiis'] = ((df.msng == 0) & (df[col1].str.contains(r'\d') == False) & (df[col2].str.contains(r'\d') == False)).astype(int)
    features_df[varname + '_missing'] = df.msng

    features_df.loc[df.msng == 0, varname + '_edit_dist'] = \
            np.vectorize(editdistance.eval, otypes=['float'])(
                df[df.msng == 0][col1].values,
                df[df.msng == 0][col2].values)

    features_df.loc[df.msng == 0, varname + '_exact_match'] = \
            (df[col1] == df[col2]).astype(float)

    features_df.loc[df.msng == 0, varname + '_exact_match_1st'] = \
            (df[col1].str.slice(0, 1) == df[col2].str.slice(0, 1)).astype(float)

    features_df.loc[df.msng == 0, varname + '_exact_match_1st2nd3rd'] = \
            (df[col1].str.slice(0, 3) == df[col2].str.slice(0, 3)).astype(float)

    df.loc[df.msng == 0, 'soundex_col_1'] = np.vectorize(jellyfish.soundex, otypes=['str'])(
            df[df.msng == 0][col1])
    df.loc[df.msng == 0, 'soundex_col_2'] = np.vectorize(jellyfish.soundex, otypes=['str'])(
            df[df.msng == 0][col2])
    features_df[varname + '_soundex'] = (df.soundex_col_1 == df.soundex_col_2).astype(int)

    df.loc[df.valid_nysiis == 1, 'nysiis_col_1'] = np.vectorize(jellyfish.nysiis, otypes=['str'])(
            df[df.valid_nysiis == 1][col1].values)
    df.loc[df.valid_nysiis == 1, 'nysiis_col_2'] = np.vectorize(jellyfish.nysiis, otypes=['str'])(
            df[df.valid_nysiis == 1][col2].values)
    features_df[varname + '_nysiis'] = (df.nysiis_col_1 == df.nysiis_col_2).astype(int)

    features_df.loc[df.msng == 0, varname + '_jw_dist'] = \
            1 - np.vectorize(distance.get_jaro_distance, otypes=['float'])(
                df[df.msng == 0][col1].values,
                df[df.msng == 0][col2].values)

    return features_df


def compare_numbers(df, varname):
    '''Compute numeric distances between two records for a particular matching field.

    Args:
        df (pd.DataFrame): side-by-side table
        varname (str): name of the all-names column to compute distance measures for

    Returns:
        pd.DataFrame: features associated with the current variable
            ====================    =======================================================
            <var>_missing           flag, 1 if either record is missing the value for this field
            <var>_num_diff          numeric difference between two field valies
            ====================    =======================================================
    '''

    df = df.copy()
    features_df = pd.DataFrame(index=df.index)
    col1 = varname + '_1'
    col2 = varname + '_2'

    features_df[varname + '_missing'] = ((df[col1] == '') | (df[col2] == '')).astype(int)

    features_df.loc[features_df[varname + '_missing'] == 0, varname + '_num_diff'] = \
            (pd.to_numeric(df[col1]) - pd.to_numeric(df[col2])).abs().astype(float)

    return features_df


@profile
def compare_categories(df, varname):
    '''Compute categorical distances between two records for a particular matching field.

    Args:
        df (pd.DataFrame): side-by-side table
        varname (str): name of the all-names column to compute distance measures for

    Returns:
        pd.DataFrame: features associated with the current variable
            ====================    =======================================================
            <var>_missing           flag, 1 if either record is missing the value for this field
            <var>_exact_match       flag, 1 if categories are identical
            <var>_partial_match     flag, 1 if any part of the categories (split by " ") match
            ====================    =======================================================
    '''

    df = df.copy()
    features_df = pd.DataFrame(index=df.index)
    col1 = varname + '_1'
    col2 = varname + '_2'

    features_df[varname + '_missing'] = ((df[col1] == '') | (df[col2] == '')).astype(int)

    features_df.loc[features_df[varname + '_missing'] == 0, varname + '_exact_match'] = \
            (df[col1] == df[col2]).astype(float)

    # features_df.loc[features_df[varname + '_missing'] == 0, varname + '_partial_match'] = \
    #         df[[col1, col2]].apply(lambda row: \
    #         int(any(col1_piece in row[col2].split(' ') for col1_piece in row[col1].split(' '))), axis=1)

    return features_df


@profile
def compare_dates(df, varname):
    '''Compute date distances between two records for a particular matching field.
    NOTE: The date MUST be in the format yyyy-mm-dd (handled upstream).

    Args:
        df (pd.DataFrame): side-by-side table
        varname (str): name of the all-names column to compute distance measures for

    Returns:
        pd.DataFrame: features associated with the current variable
            ====================    =======================================================
            <var>_missing           flag, 1 if either record is missing the value for this field
            <var>_edit_dist         number of character edits between two date values
            <var>_day_diff          number of days between two date values
            ====================    =======================================================
    '''

    df = df.copy()
    features_df = pd.DataFrame(index=df.index)
    col1 = varname + '_1'
    col2 = varname + '_2'

    df['date1'] = pd.to_datetime(df[col1], format='%Y-%m-%d', errors='coerce')
    df['date2'] = pd.to_datetime(df[col2], format='%Y-%m-%d', errors='coerce')

    features_df[varname + '_missing'] = (df.date1.isnull() | df.date2.isnull()).astype(int)

    features_df.loc[features_df[varname + '_missing'] == 0, varname + '_edit_dist'] = \
            np.vectorize(editdistance.eval, otypes=['float'])(
                df[features_df[varname + '_missing'] == 0][col1].values,
                df[features_df[varname + '_missing'] == 0][col2].values).astype(float)

    features_df.loc[features_df[varname + '_missing'] == 0, varname + '_day_diff'] = \
            ((df.date1 - df.date2) / np.timedelta64(1, 'D')).abs().astype(float)

    return features_df


def compare_geographies(df, varname):
    '''Compute geographic distance between two records for a particular matching field.

    Args:
        df (pd.DataFrame): side-by-side table
        varname (str): name of the all-names column to compute distance measures for

    Returns:
        pd.DataFrame: features associated with the current variable
            ====================    =======================================================
            <var>_missing           flag, 1 if either record is missing the value for this field
            <var>_geog_dist         euclidian distance between two geographic points
            ====================    =======================================================
    '''

    df = df.copy()
    features_df = pd.DataFrame(index=df.index)
    col1 = varname + '_1'
    col2 = varname + '_2'

    df[['x1', 'y1']] = df[col1].str.split(',', n=1, expand=True)
    df[['x2', 'y2']] = df[col2].str.split(',', n=1, expand=True)
    df[['x1', 'y1', 'x2', 'y2']] = df[['x1', 'y1', 'x2', 'y2']].replace('', np.NaN)
    df['x1_minus_x2_squared'] = np.square(df.x1.astype(float) - df.x2.astype(float))
    df['y1_minus_y2_squared'] = np.square(df.y1.astype(float) - df.y2.astype(float))

    features_df[varname + '_missing'] = ((df[col1] == '') | (df[col2] == '')).astype(int)

    features_df.loc[features_df[varname + '_missing'] == 0, varname + '_geog_dist'] = \
            np.sqrt(df.x1_minus_x2_squared + df.y1_minus_y2_squared).astype(float)

    return features_df


@profile
def generate_label(df, uid_vars, leven_thresh):
    '''Create a column indicating whether the UIDs of a pair of people is a match. The value is 1
    if UIDs are non-missing match, 0 if UIDs are non-missing and don't match, and missing if either
    UID is missing or if (0 < edit distance <= `leven_thresh`).

    The `leven_thresh` will never flip a mismatch to a match, only to a missing.

    Args:
        df (pd.DataFrame): side-by-side table
        uid_vars (list): column name(s) of the UniqueID variable(s)
        leven_thresh (int): n character edits to allow between uids, below which 0 labels will get flipped to ""

    Returns:
        pd.Series: label column
    '''

    df = df.copy()

    label_df = pd.DataFrame(index=df.index)

    for uid in uid_vars:

        col1 = uid + '_1'
        col2 = uid + '_2'

        label_df[uid + '_label'] = ''
        label_df.loc[(df[col1] != '') & (df[col2] != ''), uid + '_label'] = (df[col1] == df[col2]).astype(int)

        # if leven_thresh is in use, set missing all labels that are too close
        if leven_thresh is not None:
            label_df[uid + '_ed'] =  np.vectorize(editdistance.eval, otypes=['float'])(
                    df[col1].values, df[col2].values)
            label_df.loc[(label_df[uid + '_ed'] <= leven_thresh) &
                    (label_df[uid + '_ed'] > 0), uid + '_label'] = ''

    # num_0s is the number of times the IDs actually don't match-- a near-match would be stored as a ""
    label_df['num_1s'] = (label_df.filter(regex='_label') == 1).sum(axis=1)
    label_df['num_0s'] = (label_df.filter(regex='_label') == 0).sum(axis=1)

    label_df['label'] = ''

    label_df.loc[(label_df.num_0s == 0) & (label_df.num_1s > 0), 'label'] = '1'
    label_df.loc[(label_df.num_0s > 0) & (label_df.num_1s == 0), 'label'] = '0'
    # NOTE: if there are multiple unique ids, only use pairs that always agree as 1s;
    #       only use pairs that always disagree as 0s

    return label_df.label



import hashlib
import json
import logging.config
import logging
import os
import numpy as np
import pandas as pd
import pickle
import random
import re
import time
import yaml
import io

from collections.abc import Mapping
from functools import partial, wraps, update_wrapper

import pyarrow as pa
import pyarrow.parquet as pq

from ruamel.yaml.scalarstring import DoubleQuotedScalarString
from memory_profiler import memory_usage


class StatLogFilter():
    def __init__(self):
        self.__level = 100

    def filter(self, logRecord):
        return logRecord.levelno < self.__level


def setup_logging(log_params, log_filepath, output_temp_dir, filter_stats=False, logging_level='INFO'):
    """Setup logging configuration.

    Args:
        log_params (dict): contains info for logging setup
        log_filepath (str): path to store logs
    """

    if log_filepath is not None:
        log_params['handlers']['file_handler']['filename'] = log_filepath

    if not os.path.exists(os.path.dirname(log_filepath)):
        os.makedirs(os.path.dirname(log_filepath))

    # mappingproxy type is not hashable!
    log_params['filters']['stat_filter']['()'] = StatLogFilter

    if filter_stats:
        log_params['handlers']['file_handler']['filters'] = ['stat_filter']
        log_params['handlers']['console']['filters'] = ['stat_filter']

    runtime_num = 8
    logging.addLevelName(runtime_num, "RUNTIME")

    def runtime(self, message, *args, **kws):
        self._log(runtime_num, message, args, **kws)
    logging.Logger.runtime = runtime

    memory_num = 7
    logging.addLevelName(memory_num, "MEMORY")

    def memory(self, message, *args, **kws):
        self._log(memory_num, message, args, **kws)
    logging.Logger.memory = memory

    stat_num = 6
    logging.addLevelName(stat_num, "STAT")

    def stat(self, message, *args, **kws):
        self._log(stat_num, message, args, **kws)

    def stat_dict(self, message, *args, **kws):
        self._log(stat_num, yaml.dump(message, default_flow_style=False), args, **kws)
    logging.Logger.stat = stat
    logging.Logger.stat_dict = stat_dict

    trace_num = 5
    logging.addLevelName(trace_num, "TRACE")

    def trace(self, message, *args, **kws):
        self._log(trace_num, message, args, **kws)
    logging.Logger.trace = trace

    log_params['handlers']['console']['level'] = logging_level.upper()
    log_params['handlers']['file_handler']['level'] = logging_level.upper()
    logging.config.dictConfig(log_params)


def log_stat(human_desc, yaml_desc, value):
    '''Log a statistic in the log and in the stats yaml.

    Args:
        human_desc (str): human readable description of the stat (could be a phrase)
        yaml_desc (str): concise yaml-key compatible description of the stat
        value (float or str): value of the stat
    '''

    logger = logging.getLogger()
    logger.info(f'{human_desc}: {value}')


def log_runtime_and_memory(method):
    '''Decorator that logs time to execute functions and records max memory usage in GB.

    Args:
        method (function): function to measure/log runtime and memory usage

    Returns:
        value returned by the function being decorated
    '''
    @wraps(method)
    def inner_log_runtime_and_memory(*args, **kw):

        from datetime import timedelta
        logger = logging.getLogger()

        ts = time.time()
        memory_val, result = memory_usage(proc=(method, args, kw),
                retval=True, max_usage=True, max_iterations=1)
        te = time.time()

        elapsed_time = round(te - ts)
        elapsed_time_str = str(timedelta(seconds = elapsed_time))

        try:
            # for class method
            task = args[0]
            task_name = task.__module__.split('.')[1]
        except:
            # for normal function
            task_name = method.__module__

        if "main" in method.__name__:
            try:
                task.stats_dict[f"runtime__main__{task_name}"] = elapsed_time_str
                task.stats_dict[f"ram__main__{task_name}"] = f"{round(memory_val/1000., 2)} GB"

            except:
                raise Exception("Stats failed to record!")

        logger.runtime('%s: %s', f"runtime__{task_name}_{method.__name__}", elapsed_time_str)
        if type(memory_val) == list:
            memory_val = memory_val[0]
        logger.memory('%s: %s GB', f"memory__{task_name}_{method.__name__}", round(memory_val/1000., 2))

        return result

    return inner_log_runtime_and_memory


def load_yaml(yaml_file):
    '''Load a yaml file into a dictionary.

    Args:
        yaml_file (str): path to yaml file

    Returns:
        dict: dictionary version of input yaml file
    '''

    with open(yaml_file, 'r') as f:
        yaml_as_dict = yaml.load(f, Loader=yaml.FullLoader)

    return yaml_as_dict


def dump_yaml(dict_to_write, yaml_file):
    '''Write a dictionary into a yaml file.

    Args:
        dict_to_write (dict): dict to write to yaml
        yaml_file (str): path to output yaml file
    '''

    with open(yaml_file, 'w') as f:
        yaml.dump(dict_to_write, f, default_flow_style=False)


def to_dict(obj):
    '''Convert an object (i.e. instance of a user-defined class) into a dictionary
    to make writing easier.

    Args:
        obj (object): class instance to convert to dict
    '''
    if obj:
        return json.loads(json.dumps(obj, default=lambda o: o.__dict__))

    return {}

def create_nm_record_id(nickname, record_id_series):

    record_id_series = nickname + '__' + record_id_series
    return record_id_series


def clean_nn_string(n):
    '''Removes JR, SR, II, extra spaces, etc. from nn strings. The original string
    in the dataframe keeps punctuation and suffixes.

    Args:
        n (str): raw name value

    Returns:
        str: clean version of the input name
    '''

    # remove suffixes
    clean = re.sub("\\bSR\\b", "", n)
    clean = re.sub("\\bJR\\b", "", clean)
    clean = re.sub("\\bIII\\b", "", clean)
    clean = re.sub("\\bII\\b", "", clean)
    clean = re.sub("\\bIV\\b", "", clean)

    # drop all non letters and spaces
    regex = re.compile('[^A-Z ]')
    clean = regex.sub('', clean)

    # clean up spacing
    clean = clean.strip()
    clean = re.sub("\s+", " ", clean)
    return clean


def build_blockstring(df, blocking_scheme, incl_ed_string=True):
    '''Create blockstrings (values for blocking separated by ::, such as
    JOHN::SMITH::1993-07-23) from all-names data.

    Args:
        df (pd.DataFrame): all-names table

            =====================   =======================================================
            record_id               unique record identifier
            file_type               either "new" or "existing"
            <fields for matching>   both for the matching model and for constraint checking
            drop_from_nm            flag, 1 if met any "to drop" criteria 0 otherwise
            =====================   =======================================================

        blocking_scheme (dict): contains info about fields to block on
        incl_ed_string (bool): True if the blockstring should end with the edit-distance string (e.g. dob)

    Returns:
        pd.Series: blockstrings
    '''

    df = df.copy()

    blocking_cols = blocking_scheme['cosine_distance']['variables']
    if incl_ed_string:
        blocking_cols = \
                blocking_scheme['cosine_distance']['variables'] + \
                [blocking_scheme['edit_distance']['variable']]

    df['bs'] = ''
    for i, col in enumerate(blocking_cols):

        if i == 0:
            df['bs'] = df[col]
        else:
            df['bs'] = df.bs + '::' + df[col]

    return df['bs']


def get_nn_string_from_blockstring(blockstring):
    '''Parse out the near-neighbor string (e.g. first-name and last-name) from a blockstring.

    Args:
        blockstring (str): string with info for blocking (e.g. JOHN::SMITH::1993-07-23)

    Returns:
        str: near-neighbor string (e.g. JOHN::SMITH)
    '''

    return re.sub('::[A-Z0-9\-]*$', '', blockstring)


def get_ed_string_from_blockstring(blockstring):
    '''Parse out the edit-distance string (e.g. dob) from a blockstring.

    Args:
        blockstring (str): string with info for blocking (e.g. JOHN::SMITH::1993-07-23)

    Returns:
        str: edit-distance string (e.g. 1993-07-23)
    '''

    return re.sub('[A-Z ]*::[A-Z ]*::', '', blockstring)


def get_endpoints(n, num_chunks):
    '''Divide a number into some number of chunks/intervals.

    Args:
        n (int): number to divide into chunks/intervals
        num_chunks(int): number of chunks/intervals to create


    Returns:
        list of int tuples: list of start and end points to cover entire range
    '''

    end_points = []
    factor = int(n / num_chunks)
    for i in range(num_chunks):
        start_ix_worker = i * factor # create chunks for parallelization
        end_ix_worker = start_ix_worker + factor
        if i == num_chunks - 1:
            end_points.append([start_ix_worker, n])
        else:
            end_points.append([start_ix_worker, end_ix_worker])

    return end_points


def load_sample(csv_path, pct, cols=None):
    '''Load a random sample of a csv into pandas.

    Args:
        csv_path (str): path to csv file
        pct (float): what percent of the file to randomly read
        cols (list): columns to load

    Returns:
        pd.DataFrame: random subset of the input csv
    '''

    df_sample = pd.read_csv(csv_path, usecols=cols, header=0,
            skiprows=lambda i: i>0 and random.random() > pct)

    return df_sample


def load_csv_list(df_file_list, cols=None, conditions_dict={}, sample=1):
    '''Read a list of .csv files into a single pd.DataFrame.

    Args:
        df_file_list (list of str): list of .csv files to read
        cols (list): columns to keep in the dataframe
        conditions_dict (dict): conditions for row filtering
        sample (float): share of rows to randomly sample from the final dataframe

    Return:
        pd.DataFrame: filtered sampled dataframe read in from the .csv files
    '''

    all_df = pd.DataFrame()
    for df_file in df_file_list:

        if sample == 1:
            df = pd.read_csv(df_file, usecols=cols)
        else:
            df = load_sample(df_file, usecols=cols, pct=sample)

        for col, acceptable_value in conditions_dict.items():
            df = df[df[col] == acceptable_value]

        all_df = pd.concat([all_df, df])

    all_df = all_df.reset_index(drop=True)

    return all_df

def load_parquet(df_file, cols=None, conditions_dict={}):
    '''Read a .parquet file into a pd.DataFrame.

    Args:
        df_file (str): .parquet file to read
        cols (list): columns to keep in the dataframe
        conditions_dict (dict): conditions for row filtering

    Return:
        pd.DataFrame: filtered dataframe read in from the .parquet file
    '''
    filters = []

    if conditions_dict:
        for col, acceptable_value in conditions_dict.items():
            filters.append((col, '==', acceptable_value))
    else:
        filters = None

    pf = pq.read_table(df_file, columns=cols, filters=filters, use_threads=True)

    return pf.to_pandas()


def load_parquet_list(df_file_list, cols=None, conditions_dict={}, sample=1):
    '''Read a list of .parquet files into a single pd.DataFrame.

    Args:
        df_file_list (list of str): list of .parquet files to read
        cols (list): columns to keep in the dataframe
        conditions_dict (dict): conditions for row filtering
        sample (float): share of rows to randomly sample from the final dataframe

    Return:
        pd.DataFrame: filtered sampled dataframe read in from the .parquet files
    '''

    all_df = pd.DataFrame()
    for df_file in df_file_list:
        df = load_parquet(df_file, cols, conditions_dict)
        all_df = pd.concat([all_df, df])

    if sample == 1:
        return all_df.reset_index(drop=True)
    else:
        return all_df.sample(frac=sample).reset_index(drop=True)


def determine_model_to_use(dr_df, model_info, verbose=False):
    '''Assign a model to each data row based on which fields are available.

    Args:
        dr_df (pd.DataFrame): data rows

            ========================   ===============================================================
            record_id_1                unique identifier for the first record in the pair
            record_id_2                unique identifier for the second record in the pair
            <distance metric fields>   distance metrics between the two records' matching fields
            label                      flag, "1" if the records are a match, "0" if not, "" if unknown
            ========================   ===============================================================

        model_info (dict): information about models and their universes
        verbose (bool): flag controlling logging statement (set according to which function calls this one)

    Returns:
        pd.Series: string indicating which model to use for a given record pair
    '''

    logger = logging.getLogger()
    dr_df = dr_df.copy()

    # each data row gets assigned a model based on the values available in each field
    # TODO need to rank non-basic models if more than one and not mutually exclusive
    dr_df['model_to_use'] = 'basic'
    for this_model_name, this_model_info in model_info.items():

        # basic is the default
        if this_model_name == 'basic':
            continue

        if verbose:
            logger.debug(f'Determining rows for {this_model_name} model.')

        phat_universe = this_model_info["actual_phat_universe"]

        if len(phat_universe) == 0:
            logger.error("Special model types must have a limited actual_phat_universe.")
            raise ValueError

        acceptable = f'meets_criteria__{this_model_name}'
        dr_df[acceptable] = True
        for col, accepted_values in phat_universe.items():
            if not isinstance(accepted_values, list):
                accepted_values = [accepted_values]
            dr_df[acceptable] = (dr_df[acceptable]) & (dr_df[col].isin(accepted_values))

        dr_df.loc[dr_df[acceptable], 'model_to_use'] = this_model_name

    return dr_df['model_to_use']


def load_models(model_info_file, selection=False):
    '''Load pre-trained models (selection and match, as available)

    Args:
        model_info_file (str): path to original model config
        selection (bool): if True, try to load a corresponding selection model

    Returns:
        dict: maps model name (e.g. basic or no-dob) to a fit model object
        dict: dict with information about how to fit the model
    '''

    model_info = load_yaml(model_info_file)

    selection_models = {}
    match_models = {}
    for model_name, this_model_info in model_info.items():
        if selection:
            try:
                with open(this_model_info['selection_model_path'], 'rb') as mf:
                    selection_models[model_name] = pickle.load(mf)
            except:
                pass
        with open(this_model_info['match_model_path'], 'rb') as mf:
            match_models[model_name] = pickle.load(mf)

    if selection:
        return selection_models, match_models, model_info
    else:
        return match_models, model_info


def recursively_convert_tuple_to_list(value):
    if isinstance(value, Mapping):
        return dict((k, recursively_convert_tuple_to_list(v)) for k, v in value.items())

    elif isinstance(value, tuple):
        return list(recursively_convert_tuple_to_list(v) for v in value)
    return value


def luigi_dict_parameter_to_dict(d):
    out = {}
    for key, value in d.items():
        out[key] = recursively_convert_tuple_to_list(value)
    return out


def filename_friendly_hash(inputs):
    import datetime
    def dt_handler(x):
        if isinstance(x, datetime.datetime) or isinstance(x, datetime.date):
            return x.isoformat()
        raise TypeError("Unknown type")

    return hashlib.md5(
        json.dumps(inputs, default=dt_handler, sort_keys=True).encode("utf-8")
    ).hexdigest()


def load_logging_params(logging_params_file=None):
    default_logging_params_file = os.path.join(os.path.dirname(__file__), 'logging_config.yaml')
    logging_params_file = logging_params_file if logging_params_file else default_logging_params_file
    return yaml.load(open(logging_params_file, 'r'), Loader=yaml.FullLoader)


def reformat_dict(d: dict):
    '''make all the string values in the yaml file have double quotes'''
    d = d.copy()
    for k, v in d.items():
        if isinstance(v, str):
            d[k] = DoubleQuotedScalarString(v)
    return d


def camel_to_snake(name):
    name = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', name)
    return re.sub('([a-z0-9])([A-Z])', r'\1_\2', name).lower()
